from __future__ import annotations

import functools
import logging
import time
import typing as t
from functools import cached_property
from typing import TYPE_CHECKING

from packaging.version import parse

from ...exceptions import BentoMLException
from ...grpc.utils import LATEST_PROTOCOL_VERSION
from ...grpc.utils import import_generated_stubs
from ...grpc.utils import import_grpc
from ...grpc.utils import load_from_file
from .. import io_descriptors
from ..service import Service
from ..service.inference_api import InferenceAPI
from ..utils import LazyLoader
from . import Client

logger = logging.getLogger(__name__)

PROTOBUF_EXC_MESSAGE = "'protobuf' is required to use gRPC Client. Install with 'pip install bentoml[grpc]'."
REFLECTION_EXC_MESSAGE = "'grpcio-reflection' is required to use gRPC Client. Install with 'pip install bentoml[grpc-reflection]'."

if TYPE_CHECKING:
    import grpc
    from google.protobuf import json_format as _json_format
    from grpc import aio
    from grpc._channel import Channel as GrpcSyncChannel
    from grpc_health.v1 import health_pb2 as pb_health

    from ...grpc.v1.service_pb2 import Response
    from ...grpc.v1.service_pb2 import ServiceMetadataResponse
    from ..types import PathType

    class ClientCredentials(t.TypedDict):
        root_certificates: t.NotRequired[PathType | bytes]
        private_key: t.NotRequired[PathType | bytes]
        certificate_chain: t.NotRequired[PathType | bytes]

else:
    grpc, aio = import_grpc()
    pb_health = LazyLoader("pb_health", globals(), "grpc_health.v1.health_pb2")
    _json_format = LazyLoader(
        "_json_format",
        globals(),
        "google.protobuf.json_format",
        exc_msg=PROTOBUF_EXC_MESSAGE,
    )


# TODO: xDS support
class GrpcClient(Client):
    def __init__(
        self,
        server_url: str,
        svc: Service,
        # gRPC specific options
        ssl: bool = False,
        channel_options: aio.ChannelArgumentType | None = None,
        interceptors: t.Sequence[aio.ClientInterceptor] | None = None,
        compression: grpc.Compression | None = None,
        ssl_client_credentials: ClientCredentials | None = None,
        *,
        protocol_version: str = LATEST_PROTOCOL_VERSION,
        **kwargs: t.Any,
    ):
        self._pb, _ = import_generated_stubs(protocol_version)

        self._protocol_version = protocol_version
        self._compression = compression
        self._options = channel_options
        self._interceptors = interceptors
        self._credentials = None
        if ssl:
            assert (
                ssl_client_credentials is not None
            ), "'ssl=True' requires 'ssl_client_credentials'"
            self._credentials = grpc.ssl_channel_credentials(
                **{
                    k: load_from_file(v) if isinstance(v, str) else v
                    for k, v in ssl_client_credentials.items()
                }
            )
        self._call_rpc = f"/bentoml.grpc.{protocol_version}.BentoService/Call"
        super().__init__(svc, server_url)

    @property
    def channel(self):
        if self._credentials is not None:
            return aio.secure_channel(
                self.server_url,
                credentials=self._credentials,
                options=self._options,
                compression=self._compression,
                interceptors=self._interceptors,
            )
        else:
            return aio.insecure_channel(
                self.server_url,
                options=self._options,
                compression=self._compression,
                interceptors=self._interceptors,
            )

    @staticmethod
    def _create_sync_channel(
        server_url: str,
        ssl: bool = False,
        ssl_client_credentials: ClientCredentials | None = None,
        channel_options: t.Any | None = None,
        compression: grpc.Compression | None = None,
    ) -> GrpcSyncChannel:
        if ssl:
            assert (
                ssl_client_credentials is not None
            ), "'ssl=True' requires 'ssl_client_credentials'"
            return grpc.secure_channel(
                server_url,
                credentials=grpc.ssl_channel_credentials(
                    **{
                        k: load_from_file(v) if isinstance(v, str) else v
                        for k, v in ssl_client_credentials.items()
                    }
                ),
                options=channel_options,
                compression=compression,
            )
        return grpc.insecure_channel(
            server_url, options=channel_options, compression=compression
        )

    @staticmethod
    def wait_until_server_ready(
        host: str,
        port: int,
        timeout: float = 30,
        check_interval: int = 1,
        # set kwargs here to omit gRPC kwargs
        **kwargs: t.Any,
    ) -> None:
        protocol_version = kwargs.get("protocol_version", LATEST_PROTOCOL_VERSION)

        channel = GrpcClient._create_sync_channel(
            f"{host}:{port}",
            ssl=kwargs.get("ssl", False),
            ssl_client_credentials=kwargs.get("ssl_client_credentials", None),
            channel_options=kwargs.get("channel_options", None),
            compression=kwargs.get("compression", None),
        )
        rpc = channel.unary_unary(
            "/grpc.health.v1.Health/Check",
            request_serializer=pb_health.HealthCheckRequest.SerializeToString,
            response_deserializer=pb_health.HealthCheckResponse.FromString,
        )

        start_time = time.time()
        while time.time() - start_time < timeout:
            try:
                response = t.cast(
                    pb_health.HealthCheckResponse,
                    rpc(
                        pb_health.HealthCheckRequest(
                            service=f"bentoml.grpc.{protocol_version}.BentoService"
                        )
                    ),
                )
                if response.status == pb_health.HealthCheckResponse.SERVING:
                    break
                else:
                    time.sleep(check_interval)
            except grpc.RpcError:
                logger.debug("Server is not ready. Retrying...")
                time.sleep(check_interval)

        try:
            response = t.cast(
                pb_health.HealthCheckResponse,
                rpc(
                    pb_health.HealthCheckRequest(
                        service=f"bentoml.grpc.{protocol_version}.BentoService"
                    )
                ),
            )
            if response.status != pb_health.HealthCheckResponse.SERVING:
                raise TimeoutError(
                    f"Timed out waiting {timeout} seconds for server at '{host}:{port}' to be ready."
                )
        except (grpc.RpcError, TimeoutError) as err:
            logger.error("Caught exception while connecting to %s:%s:", host, port)
            logger.error(err)
            raise
        finally:
            channel.close()

    @cached_property
    def _rpc_metadata(self) -> dict[str, dict[str, t.Any]]:
        # Currently all RPCs in BentoService are unary-unary
        # NOTE: we will set the types of the stubs to be Any.
        return {
            method: {"input_type": input_type, "output_type": output_type}
            for method, input_type, output_type in (
                (
                    self._call_rpc,
                    self._pb.Request,
                    self._pb.Response,
                ),
                (
                    f"/bentoml.grpc.{self._protocol_version}.BentoService/ServiceMetadata",
                    self._pb.ServiceMetadataRequest,
                    self._pb.ServiceMetadataResponse,
                ),
                (
                    "/grpc.health.v1.Health/Check",
                    pb_health.HealthCheckRequest,
                    pb_health.HealthCheckResponse,
                ),
            )
        }

    async def health(self, service_name: str, *, timeout: int = 30) -> t.Any:
        return await self._invoke(
            method_name="/grpc.health.v1.Health/Check",
            service=service_name,
            _grpc_channel_timeout=timeout,
        )

    async def _invoke(self, method_name: str, **attrs: t.Any):
        # channel kwargs include timeout, metadata, credentials, wait_for_ready and compression
        # to pass it in kwargs add prefix _grpc_channel_<args>
        channel_kwargs = {
            k: attrs.pop(f"_grpc_channel_{k}", None)
            for k in {
                "timeout",
                "metadata",
                "credentials",
                "wait_for_ready",
                "compression",
            }
        }
        if method_name not in self._rpc_metadata:
            raise ValueError(
                f"'{method_name}' is a yet supported rpc. Current supported are: {self._rpc_metadata}"
            )
        metadata = self._rpc_metadata[method_name]
        rpc = self.channel.unary_unary(
            method_name,
            request_serializer=metadata["input_type"].SerializeToString,
            response_deserializer=metadata["output_type"].FromString,
        )

        return await t.cast(
            "t.Awaitable[Response]",
            rpc(metadata["input_type"](**attrs), **channel_kwargs),
        )

    async def _call(
        self,
        inp: t.Any = None,
        *,
        _bentoml_api: InferenceAPI[t.Any],
        **attrs: t.Any,
    ) -> t.Any:
        state = self.channel.get_state(try_to_connect=True)
        if state != grpc.ChannelConnectivity.READY:
            # create a blocking call to wait til channel is ready.
            await self.channel.channel_ready()

        fn = functools.partial(
            self._invoke,
            method_name=f"/bentoml.grpc.{self._protocol_version}.BentoService/Call",
            **{
                f"_grpc_channel_{k}": attrs.pop(f"_grpc_channel_{k}", None)
                for k in {
                    "timeout",
                    "metadata",
                    "credentials",
                    "wait_for_ready",
                    "compression",
                }
            },
        )

        if _bentoml_api.multi_input:
            if inp is not None:
                raise BentoMLException(
                    f"'{_bentoml_api.name}' takes multiple inputs; all inputs must be passed as keyword arguments."
                )
            serialized_req = await _bentoml_api.input.to_proto(attrs)
        else:
            serialized_req = await _bentoml_api.input.to_proto(inp)

        # A call includes api_name and given proto_fields
        api_fn = {v: k for k, v in self._svc.apis.items()}
        return await fn(
            **{
                "api_name": api_fn[_bentoml_api],
                _bentoml_api.input.proto_fields[0]: serialized_req,
            },
        )

    @classmethod
    def from_url(cls, server_url: str, **kwargs: t.Any) -> GrpcClient:
        protocol_version = kwargs.get("protocol_version", LATEST_PROTOCOL_VERSION)

        # Since v1, we introduce a ServiceMetadata rpc to retrieve bentoml.Service metadata.
        # then `client.predict` or `client.classify` won't be available.
        # client.Call will still persist for both protocol version.
        if parse(protocol_version) < parse("v1"):
            exception_message = [
                f"Using protocol version {protocol_version} older than v1. 'bentoml.client.Client' will only support protocol version v1 onwards. To create client with protocol version '{protocol_version}', do the following:\n"
                f"""\

from bentoml.grpc.utils import import_generated_stubs, import_grpc

pb, services = import_generated_stubs("{protocol_version}")

grpc, _ = import_grpc()

def run():
    with grpc.insecure_channel("localhost:3000") as channel:
        stubs = services.BentoServiceStub(channel)
        req = stubs.Call(
            request=pb.Request(
                api_name="predict",
                ndarray=pb.NDArray(
                    dtype=pb.NDArray.DTYPE_FLOAT,
                    shape=(1, 4),
                    float_values=[5.9, 3, 5.1, 1.8],
                    ),
                )
            )
            print(req)

if __name__ == '__main__':
    run()
"""
            ]
            raise BentoMLException("\n".join(exception_message))
        pb, _ = import_generated_stubs(protocol_version)

        with GrpcClient._create_sync_channel(
            server_url.replace(r"localhost", "0.0.0.0"),
            ssl=kwargs.get("ssl", False),
            ssl_client_credentials=kwargs.get("ssl_client_credentials", None),
            channel_options=kwargs.get("channel_options", None),
            compression=kwargs.get("compression", None),
        ) as channel:
            # create an insecure channel to invoke ServiceMetadata rpc
            metadata = t.cast(
                "ServiceMetadataResponse",
                channel.unary_unary(
                    f"/bentoml.grpc.{protocol_version}.BentoService/ServiceMetadata",
                    request_serializer=pb.ServiceMetadataRequest.SerializeToString,
                    response_deserializer=pb.ServiceMetadataResponse.FromString,
                )(pb.ServiceMetadataRequest()),
            )
        dummy_service = Service(metadata.name)

        for api in metadata.apis:
            try:
                dummy_service.apis[api.name] = InferenceAPI[t.Any](
                    None,
                    io_descriptors.from_spec(
                        {
                            "id": api.input.descriptor_id,
                            "args": _json_format.MessageToDict(
                                api.input.attributes
                            ).get("args", None),
                        }
                    ),
                    io_descriptors.from_spec(
                        {
                            "id": api.output.descriptor_id,
                            "args": _json_format.MessageToDict(
                                api.output.attributes
                            ).get("args", None),
                        }
                    ),
                    name=api.name,
                    doc=api.docs,
                )
            except BentoMLException as e:
                logger.error("Failed to instantiate client for API %s: ", api.name, e)

        return cls(server_url, dummy_service, **kwargs)



from __future__ import annotations

import itertools
import pickle
import typing as t
from typing import TYPE_CHECKING

from simple_di import inject

from ....exceptions import MissingDependencyException
from ...runner.container import DataContainer
from ...runner.container import DataContainerRegistry
from ...runner.container import Payload
from ...types import LazyType
from ...utils import LazyLoader

try:
    import jaxlib as jaxlib
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "jax requires jaxlib to be installed. See https://github.com/google/jax#installation for installation instructions."
    ) from None

try:
    import jax
    import jax.numpy as jnp
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "jax is required in order to use with 'bentoml.flax'. See https://github.com/google/jax#installation for installation instructions."
    ) from None

if TYPE_CHECKING:
    import numpy as np

else:
    np = LazyLoader("numpy", globals(), "numpy")

__all__ = ["jax", "jnp", "jaxlib", "JaxArrayContainer"]


class JaxArrayContainer(DataContainer[jax.Array, jax.Array]):
    @classmethod
    def batches_to_batch(
        cls,
        batches: t.Sequence[jax.Array],
        batch_dim: int = 0,
    ) -> tuple[jax.Array, list[int]]:
        batch: jax.Array = jnp.concatenate(batches, axis=batch_dim)
        indices: list[int] = list(
            itertools.accumulate(subbatch.shape[0] for subbatch in batches)
        )
        indices = [0] + indices
        return batch, indices

    @classmethod
    def batch_to_batches(
        cls,
        batch: jax.Array,
        indices: t.Sequence[int],
        batch_dim: int = 0,
    ) -> list[jax.Array]:
        return jnp.split(batch, indices[1:-1], axis=batch_dim)

    @classmethod
    @inject
    def to_payload(
        cls,
        batch: jax.Array,
        batch_dim: int = 0,
    ) -> Payload:
        return cls.create_payload(
            pickle.dumps(np.asarray(batch)),
            batch.shape[batch_dim],
        )

    @classmethod
    @inject
    def from_payload(
        cls,
        payload: Payload,
    ) -> jax.Array:
        return jnp.asarray(pickle.loads(payload.data))

    @classmethod
    @inject
    def batch_to_payloads(
        cls,
        batch: jax.Array,
        indices: t.Sequence[int],
        batch_dim: int = 0,
    ) -> t.List[Payload]:
        batches = cls.batch_to_batches(batch, indices, batch_dim)
        payloads = [cls.to_payload(subbatch, batch_dim) for subbatch in batches]
        return payloads

    @classmethod
    @inject
    def from_batch_payloads(
        cls,
        payloads: t.Sequence[Payload],
        batch_dim: int = 0,
    ) -> tuple[jax.Array, list[int]]:
        batches = [cls.from_payload(payload) for payload in payloads]
        return cls.batches_to_batch(batches, batch_dim)


DataContainerRegistry.register_container(
    LazyType("jax.numpy", "ndarray"),
    LazyType("jax.numpy", "ndarray"),
    JaxArrayContainer,
)

DataContainerRegistry.register_container(
    LazyType("jax", "Array"),
    LazyType("jax", "Array"),
    JaxArrayContainer,
)



from __future__ import annotations

import logging
import typing as t
from typing import TYPE_CHECKING

import diffusers
import torch

from .utils import get_model_or_download
from .utils import resolve_pipeline_class

if TYPE_CHECKING:
    import bentoml
    from bentoml._internal.frameworks.diffusers import LoraOptionType
    from bentoml._internal.frameworks.diffusers import TextualInversionOptionType


logger = logging.getLogger(__name__)

MODEL_NAME = "stable_diffusion"
MODEL_SHORT_NAME = "sd"
DEFAULT_MODEL_ID = "stabilityai/stable-diffusion-2-1"
PIPELINE_MAPPING = {
    "text2img": diffusers.StableDiffusionPipeline,
    "img2img": diffusers.StableDiffusionImg2ImgPipeline,
}
DEFAULT_SIZE = (768, 768)


def create_runner(
    model_id: str,
    *,
    name: str | None = None,
    use_available: bool = False,
    pipeline_class: str
    | type[diffusers.DiffusionPipeline] = diffusers.StableDiffusionPipeline,
    scheduler_class: str | type[diffusers.SchedulerMixin] | None = None,
    torch_dtype: str | torch.dtype | None = None,
    enable_xformers: bool | None = None,
    enable_attention_slicing: int | str | None = None,
    enable_model_cpu_offload: bool | None = None,
    enable_sequential_cpu_offload: bool | None = None,
    enable_torch_compile: bool | None = None,
    low_cpu_mem_usage: bool | None = None,
    variant: str | None = None,
    load_pretrained_extra_kwargs: dict[str, t.Any] | None = None,
    lora_dir: str | None = None,
    lora_weights: LoraOptionType | list[LoraOptionType] | None = None,
    textual_inversions: TextualInversionOptionType
    | list[TextualInversionOptionType]
    | None = None,
) -> bentoml.Runner:
    if isinstance(pipeline_class, str):
        pipeline_class = resolve_pipeline_class(pipeline_class, PIPELINE_MAPPING)

    model = get_model_or_download(
        MODEL_SHORT_NAME,
        model_id,
        use_available=use_available,
        pipeline_class=pipeline_class,
    )

    options = dict(
        pipeline_class=pipeline_class,
        scheduler_class=scheduler_class,
        torch_dtype=torch_dtype,
        enable_xformers=enable_xformers,
        enable_attention_slicing=enable_attention_slicing,
        enable_model_cpu_offload=enable_model_cpu_offload,
        enable_sequential_cpu_offload=enable_sequential_cpu_offload,
        enable_torch_compile=enable_torch_compile,
        low_cpu_mem_usage=low_cpu_mem_usage,
        variant=variant,
        load_pretrained_extra_kwargs=load_pretrained_extra_kwargs,
        lora_dir=lora_dir,
        lora_weights=lora_weights,
        textual_inversions=textual_inversions,
    )

    return model.with_options(**options).to_runner()



from __future__ import annotations

import functools
import json
import logging
import pickle
import typing as t
from typing import TYPE_CHECKING

from simple_di import Provide
from simple_di import inject

from bentoml.exceptions import BentoMLException
from bentoml.exceptions import ServiceUnavailable

from ..configuration.containers import BentoMLContainer
from ..context import component_context
from ..context import trace_context
from ..marshal.dispatcher import CorkDispatcher
from ..runner.container import AutoContainer
from ..runner.container import Payload
from ..runner.utils import PAYLOAD_META_HEADER
from ..runner.utils import Params
from ..runner.utils import payload_paramss_to_batch_params
from ..server.base_app import BaseAppFactory
from ..types import LazyType
from ..utils.metrics import exponential_buckets

feedback_logger = logging.getLogger("bentoml.feedback")
logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from opentelemetry.sdk.trace import Span
    from starlette.middleware import Middleware
    from starlette.requests import Request
    from starlette.responses import Response
    from starlette.routing import BaseRoute

    from ..runner.runner import Runner
    from ..runner.runner import RunnerMethod
    from ..types import LifecycleHook


class RunnerAppFactory(BaseAppFactory):
    @inject
    def __init__(
        self,
        runner: Runner,
        worker_index: int = 0,
        enable_metrics: bool = Provide[BentoMLContainer.runners_config.metrics.enabled],
    ) -> None:
        self.runner = runner
        self.worker_index = worker_index
        self.enable_metrics = enable_metrics

        self.dispatchers: dict[str, CorkDispatcher] = {}

        runners_config = BentoMLContainer.runners_config.get()
        traffic = runners_config.get("traffic", {}).copy()
        if runner.name in runners_config:
            traffic.update(runners_config[runner.name].get("traffic", {}))
        super().__init__(
            timeout=traffic["timeout"], max_concurrency=traffic["max_concurrency"]
        )

        def fallback():
            return ServiceUnavailable("process is overloaded")

        for method in runner.runner_methods:
            max_batch_size = method.max_batch_size if method.config.batchable else -1
            self.dispatchers[method.name] = CorkDispatcher(
                max_latency_in_ms=method.max_latency_ms,
                max_batch_size=max_batch_size,
                fallback=fallback,
            )

    @property
    def name(self) -> str:
        return self.runner.name

    def _init_metrics_wrappers(self):
        metrics_client = BentoMLContainer.metrics_client.get()

        max_max_batch_size = max(
            method.max_batch_size for method in self.runner.runner_methods
        )

        self.adaptive_batch_size_hist = metrics_client.Histogram(
            namespace="bentoml_runner",
            name="adaptive_batch_size",
            documentation="Runner adaptive batch size",
            labelnames=[
                "runner_name",
                "worker_index",
                "method_name",
                "service_version",
                "service_name",
            ],
            buckets=exponential_buckets(1, 2, max_max_batch_size),
        )

    @property
    def on_startup(self) -> list[LifecycleHook]:
        on_startup = super().on_startup
        on_startup.insert(0, functools.partial(self.runner.init_local, quiet=True))
        on_startup.insert(0, self._init_metrics_wrappers)

        return on_startup

    @property
    def on_shutdown(self) -> list[LifecycleHook]:
        on_shutdown: list[LifecycleHook] = [self.runner.destroy]
        for dispatcher in self.dispatchers.values():
            on_shutdown.append(dispatcher.shutdown)
        on_shutdown.extend(super().on_shutdown)
        return on_shutdown

    @property
    def routes(self) -> t.List[BaseRoute]:
        """
        Setup routes for Runner server, including:

        /healthz        liveness probe endpoint
        /readyz         Readiness probe endpoint
        /metrics        Prometheus metrics endpoint

        For method in self.runner.runner_methods:
        /{method.name}  Run corresponding runnable method
        /               Run the runnable method "__call__" if presented
        """
        from starlette.routing import Route

        routes = super().routes
        for method in self.runner.runner_methods:
            path = "/" if method.name == "__call__" else "/" + method.name
            routes.append(
                Route(
                    path=path,
                    endpoint=self._mk_request_handler(method, method.config.batchable),
                    methods=["POST"],
                )
            )
        return routes

    @property
    def middlewares(self) -> list[Middleware]:
        middlewares = super().middlewares

        from opentelemetry.instrumentation.asgi import OpenTelemetryMiddleware
        from starlette.middleware import Middleware

        def client_request_hook(span: Span, _scope: t.Dict[str, t.Any]) -> None:
            if span is not None:
                trace_context.request_id = span.context.span_id

        middlewares.append(
            Middleware(
                OpenTelemetryMiddleware,
                excluded_urls=BentoMLContainer.tracing_excluded_urls.get(),
                default_span_details=None,
                server_request_hook=None,
                client_request_hook=client_request_hook,
                tracer_provider=BentoMLContainer.tracer_provider.get(),
            )
        )

        if self.enable_metrics:
            from .http.instruments import RunnerTrafficMetricsMiddleware

            middlewares.append(Middleware(RunnerTrafficMetricsMiddleware))

        access_log_config = BentoMLContainer.runners_config.logging.access
        if access_log_config.enabled.get():
            from .http.access import AccessLogMiddleware

            access_logger = logging.getLogger("bentoml.access")
            if access_logger.getEffectiveLevel() <= logging.INFO:
                middlewares.append(
                    Middleware(
                        AccessLogMiddleware,
                        has_request_content_length=access_log_config.request_content_length.get(),
                        has_request_content_type=access_log_config.request_content_type.get(),
                        has_response_content_length=access_log_config.response_content_length.get(),
                        has_response_content_type=access_log_config.response_content_type.get(),
                    )
                )

        return middlewares

    def _mk_request_handler(
        self,
        runner_method: RunnerMethod[t.Any, t.Any, t.Any],
        batching: bool = True,
    ) -> t.Callable[[Request], t.Coroutine[None, None, Response]]:
        from starlette.responses import Response

        server_str = f"BentoML-Runner/{self.runner.name}/{runner_method.name}/{self.worker_index}"

        if runner_method.config.is_stream:
            # Streaming does not have batching implemented yet
            async def infer_stream(
                paramss: t.Sequence[Params[t.Any]],
            ) -> t.Sequence[t.AsyncGenerator[Payload, None]]:
                async def inner():
                    # This is a workaround to allow infer stream to return a iterable of
                    # async generator, to align with how non stream inference works
                    params = paramss[0].map(AutoContainer.from_payload)
                    ret = runner_method.async_stream(*params.args, **params.kwargs)
                    async for data in ret:
                        payload = AutoContainer.to_payload(data, 0)
                        yield payload

                return (inner(),)

            infer = self.dispatchers[runner_method.name](infer_stream)
        else:
            if batching:

                async def infer_batch(
                    params_list: t.Sequence[Params[t.Any]],
                ) -> list[Payload] | list[tuple[Payload, ...]]:
                    self.adaptive_batch_size_hist.labels(  # type: ignore
                        runner_name=self.runner.name,
                        worker_index=self.worker_index,
                        method_name=runner_method.name,
                        service_version=component_context.bento_version,
                        service_name=component_context.bento_name,
                    ).observe(len(params_list))

                    if not params_list:
                        return []

                    input_batch_dim, output_batch_dim = runner_method.config.batch_dim

                    batched_params, indices = payload_paramss_to_batch_params(
                        params_list, input_batch_dim
                    )

                    batch_ret = await runner_method.async_run(
                        *batched_params.args, **batched_params.kwargs
                    )

                    # multiple output branch
                    if LazyType["tuple[t.Any, ...]"](tuple).isinstance(batch_ret):
                        output_num = len(batch_ret)
                        payloadss = tuple(
                            AutoContainer.batch_to_payloads(
                                batch_ret[idx], indices, batch_dim=output_batch_dim
                            )
                            for idx in range(output_num)
                        )
                        ret = list(zip(*payloadss))
                        return ret

                    # single output branch
                    payloads = AutoContainer.batch_to_payloads(
                        batch_ret,
                        indices,
                        batch_dim=output_batch_dim,
                    )
                    return payloads

                infer = self.dispatchers[runner_method.name](infer_batch)
            else:

                async def infer_single(paramss: t.Sequence[Params[t.Any]]):
                    assert len(paramss) == 1

                    params = paramss[0].map(AutoContainer.from_payload)

                    ret = await runner_method.async_run(*params.args, **params.kwargs)

                    payload = AutoContainer.to_payload(ret, 0)
                    return (payload,)

                infer = self.dispatchers[runner_method.name](infer_single)

        async def _request_handler(request: Request) -> Response:
            assert self._is_ready

            arg_num = int(request.headers["args-number"])
            r_: bytes = await request.body()

            if arg_num == 1:
                params: Params[t.Any] = _deserialize_single_param(request, r_)

            else:
                params: Params[t.Any] = pickle.loads(r_)

            try:
                payload = await infer(params)
            except BentoMLException as e:
                # pass known exceptions to the client
                return Response(
                    status_code=e.error_code,
                    content=str(e),
                    headers={
                        PAYLOAD_META_HEADER: "{}",
                        "Content-Type": "application/vnd.bentoml.error",
                        "Server": server_str,
                    },
                )

            if isinstance(payload, ServiceUnavailable):
                return Response(
                    "Service Busy",
                    status_code=payload.error_code,
                    headers={
                        PAYLOAD_META_HEADER: json.dumps({}),
                        "Content-Type": "application/vnd.bentoml.error",
                        "Server": server_str,
                    },
                )
            if isinstance(payload, Payload):
                return Response(
                    payload.data,
                    headers={
                        PAYLOAD_META_HEADER: json.dumps(payload.meta),
                        "Content-Type": f"application/vnd.bentoml.{payload.container}",
                        "Server": server_str,
                    },
                )
            if isinstance(payload, tuple):
                # a tuple, which means user runnable has multiple outputs
                return Response(
                    pickle.dumps(payload),
                    headers={
                        PAYLOAD_META_HEADER: json.dumps({}),
                        "Content-Type": "application/vnd.bentoml.multiple_outputs",
                        "Server": server_str,
                    },
                )
            from starlette.responses import StreamingResponse

            if runner_method.config.is_stream:

                async def streamer() -> t.AsyncGenerator[bytes, None]:
                    """
                    Extract Data from a AsyncGenerator[Payload, None]
                    """
                    async for p in payload:
                        yield pickle.dumps(p)

                return StreamingResponse(
                    streamer(),
                    media_type="text/plain",
                    headers={
                        PAYLOAD_META_HEADER: json.dumps({}),
                        "Content-Type": "application/vnd.bentoml.stream_outputs",
                        "Server": server_str,
                    },
                )

            raise BentoMLException(
                f"Unexpected payload type: {type(payload)}, {payload}"
            )

        return _request_handler


def _deserialize_single_param(request: Request, bs: bytes) -> Params[t.Any]:
    container = request.headers["Payload-Container"]
    meta = json.loads(request.headers["Payload-Meta"])
    batch_size = int(request.headers["Batch-Size"])
    kwarg_name = request.headers.get("Kwarg-Name")
    payload = Payload(
        data=bs,
        meta=meta,
        batch_size=batch_size,
        container=container,
    )
    if kwarg_name:
        d = {kwarg_name: payload}
        params: Params[t.Any] = Params(**d)
    else:
        params: Params[t.Any] = Params(payload)

    return params



from __future__ import annotations

import abc
import base64
import io
import itertools
import pickle
import typing as t

from ..types import LazyType
from ..utils import LazyLoader
from ..utils.pickle import fixed_torch_loads
from ..utils.pickle import pep574_dumps
from ..utils.pickle import pep574_loads

SingleType = t.TypeVar("SingleType")
BatchType = t.TypeVar("BatchType")

TRITON_EXC_MSG = "tritonclient is required to use triton with BentoML. Install with 'pip install \"tritonclient[all]>=2.29.0\"'."

if t.TYPE_CHECKING:
    import tritonclient.grpc.aio as tritongrpcclient
    import tritonclient.http.aio as tritonhttpclient

    from .. import external_typing as ext
else:
    np = LazyLoader("np", globals(), "numpy")
    tritongrpcclient = LazyLoader(
        "tritongrpcclient", globals(), "tritonclient.grpc.aio", exc_msg=TRITON_EXC_MSG
    )
    tritonhttpclient = LazyLoader(
        "tritonhttpclient", globals(), "tritonclient.http.aio", exc_msg=TRITON_EXC_MSG
    )


class Payload(t.NamedTuple):
    data: bytes
    meta: dict[str, bool | int | float | str | list[int]]
    container: str
    batch_size: int = -1


class DataContainer(t.Generic[SingleType, BatchType]):
    @classmethod
    def create_payload(
        cls,
        data: bytes,
        batch_size: int,
        meta: dict[str, bool | int | float | str | list[int]] | None = None,
    ) -> Payload:
        return Payload(data, meta or {}, container=cls.__name__, batch_size=batch_size)

    @classmethod
    @abc.abstractmethod
    def to_payload(cls, batch: BatchType, batch_dim: int) -> Payload:
        ...

    @classmethod
    @abc.abstractmethod
    def from_payload(cls, payload: Payload) -> BatchType:
        ...

    @t.overload
    @classmethod
    def to_triton_payload(
        cls,
        inp: tritongrpcclient.InferInput,
        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
        _use_http_client: t.Literal[False] = ...,
    ) -> tritongrpcclient.InferInput:
        ...

    @t.overload
    @classmethod
    def to_triton_payload(
        cls,
        inp: tritonhttpclient.InferInput,
        meta: dict[str, t.Any],
        _use_http_client: t.Literal[True] = ...,
    ) -> tritongrpcclient.InferInput:
        ...

    @classmethod
    def to_triton_payload(
        cls, inp: SingleType, meta: t.Any, _use_http_client: bool = False
    ) -> t.Any:
        """
        Convert given input types to a Triton payload.

        Implementation of this method should always return a InferInput that has
        data of a Numpy array.
        """
        if _use_http_client:
            return cls.to_triton_http_payload(inp, meta)
        else:
            return cls.to_triton_grpc_payload(inp, meta)

    @classmethod
    def to_triton_grpc_payload(
        cls,
        inp: SingleType,
        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
    ) -> tritongrpcclient.InferInput:
        """
        Convert given input types to a Triton payload via gRPC client.

        Implementation of this method should always return a InferInput.
        """
        raise NotImplementedError(
            f"{cls.__name__} doesn't support converting to Triton payload."
        )

    @classmethod
    def to_triton_http_payload(
        cls,
        inp: SingleType,
        meta: dict[str, t.Any],
    ) -> tritonhttpclient.InferInput:
        """
        Convert given input types to a Triton payload via HTTP client.

        Implementation of this method should always return a InferInput.
        """
        raise NotImplementedError(
            f"{cls.__name__} doesn't support converting to Triton payload."
        )

    @classmethod
    @abc.abstractmethod
    def batches_to_batch(
        cls, batches: t.Sequence[BatchType], batch_dim: int
    ) -> tuple[BatchType, list[int]]:
        ...

    @classmethod
    @abc.abstractmethod
    def batch_to_batches(
        cls, batch: BatchType, indices: t.Sequence[int], batch_dim: int
    ) -> list[BatchType]:
        ...

    @classmethod
    @abc.abstractmethod
    def batch_to_payloads(
        cls, batch: BatchType, indices: t.Sequence[int], batch_dim: int
    ) -> list[Payload]:
        ...

    @classmethod
    @abc.abstractmethod
    def from_batch_payloads(
        cls, payloads: t.Sequence[Payload], batch_dim: int
    ) -> tuple[BatchType, list[int]]:
        ...


class TritonInferInputDataContainer(
    DataContainer[
        t.Union["tritongrpcclient.InferInput", "tritonhttpclient.InferInput"],
        t.Union["tritongrpcclient.InferInput", "tritonhttpclient.InferInput"],
    ]
):
    @classmethod
    def to_payload(cls, batch: tritongrpcclient.InferInput, batch_dim: int) -> Payload:
        raise NotImplementedError

    @classmethod
    def from_payload(cls, payload: Payload) -> tritongrpcclient.InferInput:
        raise NotImplementedError

    @classmethod
    def to_triton_grpc_payload(
        cls,
        inp: tritongrpcclient.InferInput | tritonhttpclient.InferInput,
        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
    ) -> tritongrpcclient.InferInput:
        return t.cast("tritongrpcclient.InferInput", inp)

    @classmethod
    def to_triton_http_payload(
        cls,
        inp: tritongrpcclient.InferInput | tritonhttpclient.InferInput,
        meta: dict[str, t.Any],
    ) -> tritonhttpclient.InferInput:
        return t.cast("tritonhttpclient.InferInput", inp)

    @classmethod
    @abc.abstractmethod
    def batches_to_batch(
        cls,
        batches: t.Sequence[tritongrpcclient.InferInput | tritonhttpclient.InferInput],
        batch_dim: int,
    ) -> tuple[tritongrpcclient.InferInput | tritonhttpclient.InferInput, list[int]]:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def batch_to_batches(
        cls,
        batch: tritongrpcclient.InferInput | tritonhttpclient.InferInput,
        indices: t.Sequence[int],
        batch_dim: int,
    ) -> list[tritongrpcclient.InferInput | tritonhttpclient.InferInput]:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def batch_to_payloads(
        cls,
        batch: tritongrpcclient.InferInput | tritonhttpclient.InferInput,
        indices: t.Sequence[int],
        batch_dim: int,
    ) -> list[Payload]:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_batch_payloads(
        cls,
        payloads: t.Sequence[Payload],
        batch_dim: int,
    ) -> tuple[tritongrpcclient.InferInput | tritonhttpclient.InferInput, list[int]]:
        raise NotImplementedError


class NdarrayContainer(DataContainer["ext.NpNDArray", "ext.NpNDArray"]):
    @classmethod
    def batches_to_batch(
        cls,
        batches: t.Sequence[ext.NpNDArray],
        batch_dim: int = 0,
    ) -> tuple[ext.NpNDArray, list[int]]:
        # numpy.concatenate may consume lots of memory, need optimization later
        batch: ext.NpNDArray = np.concatenate(batches, axis=batch_dim)
        indices = list(
            itertools.accumulate(subbatch.shape[batch_dim] for subbatch in batches)
        )
        indices = [0] + indices
        return batch, indices

    @classmethod
    def batch_to_batches(
        cls,
        batch: ext.NpNDArray,
        indices: t.Sequence[int],
        batch_dim: int = 0,
    ) -> list[ext.NpNDArray]:
        return np.split(batch, indices[1:-1], axis=batch_dim)

    @classmethod
    def to_triton_grpc_payload(
        cls,
        inp: ext.NpNDArray,
        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
    ) -> tritongrpcclient.InferInput:
        InferInput = tritongrpcclient.InferInput(
            meta.name, inp.shape, tritongrpcclient.np_to_triton_dtype(inp.dtype)
        )
        InferInput.set_data_from_numpy(inp)
        return InferInput

    @classmethod
    def to_triton_http_payload(
        cls,
        inp: ext.NpNDArray,
        meta: dict[str, t.Any],
    ) -> tritonhttpclient.InferInput:
        InferInput = tritonhttpclient.InferInput(
            meta["name"], inp.shape, tritonhttpclient.np_to_triton_dtype(inp.dtype)
        )
        InferInput.set_data_from_numpy(inp)
        return InferInput

    @classmethod
    def to_payload(
        cls,
        batch: ext.NpNDArray,
        batch_dim: int,
    ) -> Payload:
        # skip 0-dimensional array
        if batch.shape:
            if not (batch.flags["C_CONTIGUOUS"] or batch.flags["F_CONTIGUOUS"]):
                # TODO: use fortan contiguous if it's faster
                batch = np.ascontiguousarray(batch)

            bs: bytes
            concat_buffer_bs: bytes
            indices: list[int]
            bs, concat_buffer_bs, indices = pep574_dumps(batch)
            bs_str = base64.b64encode(bs).decode("ascii")

            return cls.create_payload(
                concat_buffer_bs,
                batch.shape[batch_dim],
                {
                    "format": "pickle5",
                    "pickle_bytes_str": bs_str,
                    "indices": indices,
                },
            )

        return cls.create_payload(
            pickle.dumps(batch),
            batch.shape[batch_dim],
            {"format": "default"},
        )

    @classmethod
    def from_payload(
        cls,
        payload: Payload,
    ) -> ext.NpNDArray:
        format = payload.meta.get("format", "default")
        if format == "pickle5":
            bs_str = t.cast(str, payload.meta["pickle_bytes_str"])
            bs = base64.b64decode(bs_str)
            indices = t.cast(t.List[int], payload.meta["indices"])
            return t.cast("ext.NpNDArray", pep574_loads(bs, payload.data, indices))

        return pickle.loads(payload.data)

    @classmethod
    def batch_to_payloads(
        cls,
        batch: ext.NpNDArray,
        indices: t.Sequence[int],
        batch_dim: int = 0,
    ) -> list[Payload]:
        batches = cls.batch_to_batches(batch, indices, batch_dim)

        payloads = [cls.to_payload(subbatch, batch_dim) for subbatch in batches]
        return payloads

    @classmethod
    def from_batch_payloads(
        cls,
        payloads: t.Sequence[Payload],
        batch_dim: int = 0,
    ) -> t.Tuple["ext.NpNDArray", list[int]]:
        batches = [cls.from_payload(payload) for payload in payloads]
        return cls.batches_to_batch(batches, batch_dim)


class PandasDataFrameContainer(
    DataContainer[t.Union["ext.PdDataFrame", "ext.PdSeries"], "ext.PdDataFrame"]
):
    @classmethod
    def batches_to_batch(
        cls,
        batches: t.Sequence[ext.PdDataFrame],
        batch_dim: int = 0,
    ) -> tuple[ext.PdDataFrame, list[int]]:
        import pandas as pd

        assert (
            batch_dim == 0
        ), "PandasDataFrameContainer does not support batch_dim other than 0"
        indices = list(
            itertools.accumulate(subbatch.shape[batch_dim] for subbatch in batches)
        )
        indices = [0] + indices
        return pd.concat(batches, ignore_index=True), indices  # type: ignore (incomplete panadas types)

    @classmethod
    def batch_to_batches(
        cls,
        batch: ext.PdDataFrame,
        indices: t.Sequence[int],
        batch_dim: int = 0,
    ) -> list[ext.PdDataFrame]:
        assert (
            batch_dim == 0
        ), "PandasDataFrameContainer does not support batch_dim other than 0"

        return [
            batch.iloc[indices[i] : indices[i + 1]].reset_index(drop=True)
            for i in range(len(indices) - 1)
        ]

    @classmethod
    def to_payload(
        cls,
        batch: ext.PdDataFrame | ext.PdSeries,
        batch_dim: int,
    ) -> Payload:
        import pandas as pd

        assert (
            batch_dim == 0
        ), "PandasDataFrameContainer does not support batch_dim other than 0"

        if isinstance(batch, pd.Series):
            batch = pd.DataFrame([batch])

        meta: dict[str, bool | int | float | str | list[int]] = {"format": "pickle5"}

        bs: bytes
        concat_buffer_bs: bytes
        indices: list[int]
        bs, concat_buffer_bs, indices = pep574_dumps(batch)

        if indices:
            meta["with_buffer"] = True
            data = concat_buffer_bs
            meta["pickle_bytes_str"] = base64.b64encode(bs).decode("ascii")
            meta["indices"] = indices
        else:
            meta["with_buffer"] = False
            data = bs

        return cls.create_payload(
            data,
            batch.shape[0],
            meta=meta,
        )

    @classmethod
    def from_payload(
        cls,
        payload: Payload,
    ) -> ext.PdDataFrame:
        if payload.meta["with_buffer"]:
            bs_str = t.cast(str, payload.meta["pickle_bytes_str"])
            bs = base64.b64decode(bs_str)
            indices = t.cast(t.List[int], payload.meta["indices"])
            return pep574_loads(bs, payload.data, indices)
        else:
            return pep574_loads(payload.data, b"", [])

    @classmethod
    def batch_to_payloads(
        cls,
        batch: ext.PdDataFrame,
        indices: t.Sequence[int],
        batch_dim: int = 0,
    ) -> list[Payload]:
        batches = cls.batch_to_batches(batch, indices, batch_dim)

        payloads = [cls.to_payload(subbatch, batch_dim) for subbatch in batches]
        return payloads

    @classmethod
    def from_batch_payloads(  # pylint: disable=arguments-differ
        cls,
        payloads: t.Sequence[Payload],
        batch_dim: int = 0,
    ) -> tuple[ext.PdDataFrame, list[int]]:
        batches = [cls.from_payload(payload) for payload in payloads]
        return cls.batches_to_batch(batches, batch_dim)


class PILImageContainer(DataContainer["ext.PILImage", "ext.PILImage"]):
    _error = (
        "PIL.Image doesn't support batch inference."
        "You can convert it to numpy.ndarray before passing to the runner."
    )

    @classmethod
    def to_payload(cls, batch: ext.PILImage, batch_dim: int) -> Payload:
        buffer = io.BytesIO()
        batch.save(buffer, format=batch.format)
        return cls.create_payload(buffer.getvalue(), batch_size=1)

    @classmethod
    def from_payload(cls, payload: Payload) -> ext.PILImage:
        from ..io_descriptors.image import PIL

        return PIL.Image.open(io.BytesIO(payload.data))

    @classmethod
    def batch_to_payloads(
        cls, batch: ext.PILImage, indices: t.Sequence[int], batch_dim: int
    ) -> list[Payload]:
        raise NotImplementedError(cls._error)

    @classmethod
    def batches_to_batch(
        cls, batches: t.Sequence[ext.PILImage], batch_dim: int
    ) -> tuple[ext.PILImage, list[int]]:
        raise NotImplementedError(cls._error)

    @classmethod
    def batch_to_batches(
        cls, batch: ext.PILImage, indices: t.Sequence[int], batch_dim: int
    ) -> list[ext.PILImage]:
        raise NotImplementedError(cls._error)

    @classmethod
    def from_batch_payloads(
        cls, payloads: t.Sequence[Payload], batch_dim: int = 0
    ) -> tuple[ext.PILImage, list[int]]:
        raise NotImplementedError(cls._error)


class DefaultContainer(DataContainer[t.Any, t.List[t.Any]]):
    @classmethod
    def batches_to_batch(
        cls, batches: t.Sequence[list[t.Any]], batch_dim: int = 0
    ) -> tuple[list[t.Any], list[int]]:
        assert (
            batch_dim == 0
        ), "Default Runner DataContainer does not support batch_dim other than 0"
        batch: list[t.Any] = []
        for subbatch in batches:
            batch.extend(subbatch)
        indices = list(itertools.accumulate(len(subbatch) for subbatch in batches))
        indices = [0] + indices
        return batch, indices

    @classmethod
    def batch_to_batches(
        cls, batch: list[t.Any], indices: t.Sequence[int], batch_dim: int = 0
    ) -> list[list[t.Any]]:
        assert (
            batch_dim == 0
        ), "Default Runner DataContainer does not support batch_dim other than 0"
        return [batch[indices[i] : indices[i + 1]] for i in range(len(indices) - 1)]

    @classmethod
    def to_payload(cls, batch: t.Any, batch_dim: int) -> Payload:
        if isinstance(batch, t.Generator):  # Generators can't be pickled
            batch = list(t.cast(t.Generator[t.Any, t.Any, t.Any], batch))

        data = pickle.dumps(batch)

        if isinstance(batch, list):
            batch_size = len(t.cast(t.List[t.Any], batch))
        else:
            batch_size = 1

        return cls.create_payload(data=data, batch_size=batch_size)

    @classmethod
    def from_payload(cls, payload: Payload) -> t.Any:
        return fixed_torch_loads(payload.data)

    @classmethod
    def batch_to_payloads(
        cls,
        batch: list[t.Any],
        indices: t.Sequence[int],
        batch_dim: int = 0,
    ) -> list[Payload]:
        batches = cls.batch_to_batches(batch, indices, batch_dim)

        payloads = [cls.to_payload(subbatch, batch_dim) for subbatch in batches]
        return payloads

    @classmethod
    def from_batch_payloads(
        cls,
        payloads: t.Sequence[Payload],
        batch_dim: int = 0,
    ) -> tuple[list[t.Any], list[int]]:
        batches = [cls.from_payload(payload) for payload in payloads]
        return cls.batches_to_batch(batches, batch_dim)


class DataContainerRegistry:
    CONTAINER_SINGLE_TYPE_MAP: t.Dict[
        LazyType[t.Any], t.Type[DataContainer[t.Any, t.Any]]
    ] = dict()
    CONTAINER_BATCH_TYPE_MAP: t.Dict[
        LazyType[t.Any], type[DataContainer[t.Any, t.Any]]
    ] = dict()

    @classmethod
    def register_container(
        cls,
        single_type: LazyType[t.Any] | type,
        batch_type: LazyType[t.Any] | type,
        container_cls: t.Type[DataContainer[t.Any, t.Any]],
    ):
        single_type = LazyType.from_type(single_type)
        batch_type = LazyType.from_type(batch_type)

        cls.CONTAINER_BATCH_TYPE_MAP[batch_type] = container_cls
        cls.CONTAINER_SINGLE_TYPE_MAP[single_type] = container_cls

    @classmethod
    def find_by_single_type(
        cls, type_: t.Type[SingleType] | LazyType[SingleType]
    ) -> t.Type[DataContainer[SingleType, t.Any]]:
        typeref = LazyType.from_type(type_)
        if typeref in cls.CONTAINER_SINGLE_TYPE_MAP:
            return cls.CONTAINER_SINGLE_TYPE_MAP[typeref]
        for klass in cls.CONTAINER_SINGLE_TYPE_MAP:
            if klass.issubclass(type_):
                return cls.CONTAINER_SINGLE_TYPE_MAP[klass]
        return DefaultContainer

    @classmethod
    def find_by_batch_type(
        cls, type_: t.Type[BatchType] | LazyType[BatchType]
    ) -> t.Type[DataContainer[t.Any, BatchType]]:
        typeref = LazyType.from_type(type_)
        if typeref in cls.CONTAINER_BATCH_TYPE_MAP:
            return cls.CONTAINER_BATCH_TYPE_MAP[typeref]
        for klass in cls.CONTAINER_BATCH_TYPE_MAP:
            if klass.issubclass(type_):
                return cls.CONTAINER_BATCH_TYPE_MAP[klass]
        return DefaultContainer

    @classmethod
    def find_by_name(cls, name: str) -> t.Type[DataContainer[t.Any, t.Any]]:
        for container_cls in cls.CONTAINER_BATCH_TYPE_MAP.values():
            if container_cls.__name__ == name:
                return container_cls
        if name == DefaultContainer.__name__:
            return DefaultContainer
        raise ValueError(f"can not find specified container class by name {name}")


def register_builtin_containers():
    DataContainerRegistry.register_container(
        LazyType("numpy", "ndarray"), LazyType("numpy", "ndarray"), NdarrayContainer
    )
    DataContainerRegistry.register_container(
        LazyType("pandas.core.series", "Series"),
        LazyType("pandas.core.frame", "DataFrame"),
        PandasDataFrameContainer,
    )
    DataContainerRegistry.register_container(
        LazyType("pandas.core.frame", "DataFrame"),
        LazyType("pandas.core.frame", "DataFrame"),
        PandasDataFrameContainer,
    )

    DataContainerRegistry.register_container(
        LazyType("tritonclient.grpc", "InferInput"),
        LazyType("tritonclient.grpc", "InferInput"),
        TritonInferInputDataContainer,
    )
    DataContainerRegistry.register_container(
        LazyType("tritonclient.grpc.aio", "InferInput"),
        LazyType("tritonclient.grpc.aio", "InferInput"),
        TritonInferInputDataContainer,
    )
    DataContainerRegistry.register_container(
        LazyType("tritonclient.http", "InferInput"),
        LazyType("tritonclient.http", "InferInput"),
        TritonInferInputDataContainer,
    )
    DataContainerRegistry.register_container(
        LazyType("tritonclient.http.aio", "InferInput"),
        LazyType("tritonclient.http.aio", "InferInput"),
        TritonInferInputDataContainer,
    )
    DataContainerRegistry.register_container(
        LazyType("PIL.Image", "Image"),
        LazyType("PIL.Image", "Image"),
        PILImageContainer,
    )


register_builtin_containers()


class AutoContainer(DataContainer[t.Any, t.Any]):
    @classmethod
    def to_payload(cls, batch: t.Any, batch_dim: int) -> Payload:
        container_cls: t.Type[
            DataContainer[t.Any, t.Any]
        ] = DataContainerRegistry.find_by_batch_type(type(batch))
        return container_cls.to_payload(batch, batch_dim)

    @classmethod
    def from_payload(cls, payload: Payload) -> t.Any:
        container_cls = DataContainerRegistry.find_by_name(payload.container)
        return container_cls.from_payload(payload)

    @t.overload
    @classmethod
    def to_triton_payload(
        cls,
        inp: tritongrpcclient.InferInput,
        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
        _use_http_client: t.Literal[False] = ...,
    ) -> tritongrpcclient.InferInput:
        ...

    @t.overload
    @classmethod
    def to_triton_payload(
        cls,
        inp: tritonhttpclient.InferInput,
        meta: dict[str, t.Any],
        _use_http_client: t.Literal[True] = ...,
    ) -> tritongrpcclient.InferInput:
        ...

    @classmethod
    def to_triton_payload(
        cls, inp: t.Any, meta: t.Any, _use_http_client: bool = False
    ) -> t.Any:
        """
        Convert given input types to a Triton payload.

        Implementation of this method should always return a InferInput that has
        data of a Numpy array.
        """
        if _use_http_client:
            return DataContainerRegistry.find_by_single_type(
                type(inp)
            ).to_triton_http_payload(inp, meta)
        else:
            return DataContainerRegistry.find_by_single_type(
                type(inp)
            ).to_triton_grpc_payload(inp, meta)

    @classmethod
    def batches_to_batch(
        cls, batches: t.Sequence[BatchType], batch_dim: int = 0
    ) -> tuple[BatchType, list[int]]:
        container_cls: t.Type[
            DataContainer[t.Any, t.Any]
        ] = DataContainerRegistry.find_by_batch_type(type(batches[0]))
        return container_cls.batches_to_batch(batches, batch_dim)

    @classmethod
    def batch_to_batches(
        cls, batch: BatchType, indices: t.Sequence[int], batch_dim: int = 0
    ) -> list[BatchType]:
        container_cls: t.Type[
            DataContainer[t.Any, t.Any]
        ] = DataContainerRegistry.find_by_batch_type(type(batch))
        return container_cls.batch_to_batches(batch, indices, batch_dim)

    @classmethod
    def batch_to_payloads(
        cls,
        batch: t.Any,
        indices: t.Sequence[int],
        batch_dim: int = 0,
    ) -> list[Payload]:
        container_cls: t.Type[
            DataContainer[t.Any, t.Any]
        ] = DataContainerRegistry.find_by_batch_type(type(batch))
        return container_cls.batch_to_payloads(batch, indices, batch_dim)

    @classmethod
    def from_batch_payloads(
        cls, payloads: t.Sequence[Payload], batch_dim: int = 0
    ) -> tuple[t.Any, list[int]]:
        container_cls = DataContainerRegistry.find_by_name(payloads[0].container)
        return container_cls.from_batch_payloads(payloads, batch_dim)



from ._internal.frameworks.catboost import (
    CatBoostOptions as ModelOptions,  # type: ignore # noqa
)
from ._internal.frameworks.catboost import get
from ._internal.frameworks.catboost import get_runnable
from ._internal.frameworks.catboost import load_model
from ._internal.frameworks.catboost import save_model

__all__ = ["load_model", "save_model", "get", "get_runnable"]



from __future__ import annotations

import io
import pickle
import typing as t

# Pickle protocol 5 with out-of-band data. ref: https://peps.python.org/pep-0574/

# This is originally intended for numpy ndarray/pandas dataframe
# serialization.  In these situations the `main_bytes` part will only
# contain some metadata. That's why putting these bytes in header will
# not cause trouble. In the meantime the `concat_buffer_bytes`
# contains out-of-band buffers that need no computation during
# deserialization, which will save computation resource especially
# when the payload is large. However DO NOT use this pair of functions
# on `DefaultContainer`. Because `DefaultContainer` may be a
# dictionary containing a large object that has no out-of-band buffer
# (e.g. PIL Image or PyTorch tensor) and a small numpy ndarray. In
# that case `concat_buffer_bytes` will be small and `main_bytes` may
# be huge, hence we barely save any time while having a large header
# (that may cause errors).


def pep574_dumps(obj: t.Any) -> tuple[bytes, bytes, list[int]]:
    buffers: list[pickle.PickleBuffer] = []
    main_bytes: bytes = pickle.dumps(obj, protocol=5, buffer_callback=buffers.append)

    if not buffers:
        return main_bytes, b"", []

    buffer_bytess: list[bytes] = [buff.raw().tobytes() for buff in buffers]

    for buff in buffers:
        buff.release()

    indices: list[int] = [0]
    for buff_bytes in buffer_bytess:
        start = indices[-1]
        end = start + len(buff_bytes)
        indices.append(end)

    concat_buffer_bytes: bytes = b"".join(buffer_bytess)
    return main_bytes, concat_buffer_bytes, indices


def pep574_loads(
    main_bytes: bytes, concat_buffer_bytes: bytes, indices: list[int]
) -> t.Any:
    if not indices:
        # TODO: @larme monitor
        # https://github.com/pytorch/pytorch/issues/102977 and may change
        # this function later
        return fixed_torch_loads(main_bytes)

    mem = memoryview(concat_buffer_bytes)
    partitions = zip(indices, indices[1:])
    recover_buffers: list[pickle.PickleBuffer] = []
    for partition in partitions:
        buff = pickle.PickleBuffer(mem[slice(*partition)])
        recover_buffers.append(buff)

    return fixed_torch_loads(main_bytes, buffers=recover_buffers)


def _safe_torch_tensor_loads(bs: bytes) -> t.Any:
    import torch

    f = io.BytesIO(bs)
    if not torch.cuda.is_available():
        return torch.load(f, map_location="cpu")
    else:
        return torch.load(f)


class FixTorchUnpickler(pickle.Unpickler):
    def find_class(self, module: str, name: str) -> t.Callable[[bytes], t.Any]:
        if module == "torch.storage" and name == "_load_from_bytes":
            return _safe_torch_tensor_loads
        else:
            return super().find_class(module, name)


def fixed_torch_loads(bs: bytes, **kwargs: t.Any) -> t.Any:
    f = io.BytesIO(bs)
    unpickler = FixTorchUnpickler(f, **kwargs)
    return unpickler.load()


def loads_or_fix_torch(bs: bytes):
    try:
        return pickle.loads(bs)
    except RuntimeError:
        return fixed_torch_loads(bs)



from __future__ import annotations

import importlib.metadata
import os
import platform
import shlex
import shutil
import subprocess
import sys
import typing as t
from gettext import gettext

import click

conda_packages_name = "conda_packages"

_ENVVAR = [
    "BENTOML_DEBUG",
    "BENTOML_QUIET",
    "BENTOML_BUNDLE_LOCAL_BUILD",
    "BENTOML_DO_NOT_TRACK",
    "BENTOML_CONFIG",
    "BENTOML_CONFIG_OPTIONS",
    "BENTOML_PORT",
    "BENTOML_HOST",
    "BENTOML_API_WORKERS",
]
_CONDITIONAL_ENVVAR = [
    # Only print if set
    "BENTOML_RETRY_RUNNER_REQUESTS",
    "BENTOML_CONTAINERIZE_BACKEND",
    "BENTOML_NUM_THREAD",
    "OMP_NUM_THREADS",
    "OPENBLAS_NUM_THREADS",
    "MKL_NUM_THREADS",
    "VECLIB_MAXIMUM_THREADS",
    "NUMEXPR_NUM_THREADS",
    "RAYON_RS_NUM_CPUS",
    "TF_NUM_INTEROP_THREADS",
    "TF_NUM_INTRAOP_THREADS",
    "CUDA_VISIBLE_DEVICES",
]


def run_cmd(cmd: list[str]) -> list[str]:
    return subprocess.check_output(cmd).decode("utf-8").split("\n")[:-1]


def _format_dropdown(title: str, content: t.Iterable[str]) -> str:
    processed = "\n".join(content)
    return f"""\
<details><summary><code>{title}</code></summary>

<br>

```{ 'yaml' if title == conda_packages_name else '' }
{processed}
```

</details>
"""


def _format_env(env: list[str]) -> list[str]:
    return [f"{e}={shlex.quote(os.environ.get(e, ''))}" for e in env]


def format_md(env: list[str], info_dict: dict[str, str | list[str]]) -> list[str]:
    out: list[str] = []
    out.append(
        """\
#### Environment variable

```bash
{env}
```
""".format(
            env="\n".join(_format_env(env))
        )
    )
    out.append("#### System information\n")
    for key, value in info_dict.items():
        if isinstance(value, list):
            out.append(_format_dropdown(key, value))
        else:
            out.append(f"`{key}`: {value}")
    return out


def format_bash(env: list[str], info_dict: dict[str, str | list[str]]) -> list[str]:
    out: list[str] = []
    out.extend(_format_env(env))
    for key, value in info_dict.items():
        if isinstance(value, list):
            out.append(f"{key}=( " + " ".join(map(lambda s: f'"{s}"', value)) + " )")
        else:
            out.append(f'{key}="{value}"')
    return out


def pretty_format(
    info_dict: dict[str, str | list[str]], output: t.Literal["md", "bash"]
) -> str:
    env = _ENVVAR + list(filter(lambda e: e in os.environ, _CONDITIONAL_ENVVAR))
    return "\n".join({"md": format_md, "bash": format_bash}[output](env, info_dict))


def add_env_command(cli: click.Group) -> None:
    from bentoml._internal.utils.pkg import PackageNotFoundError
    from bentoml._internal.utils.pkg import get_pkg_version
    from bentoml.exceptions import CLIException

    @cli.command(help=gettext("Print environment info and exit"))
    @click.option(
        "-o",
        "--output",
        type=click.Choice(["md", "bash"]),
        default="md",
        show_default=True,
        help="Output format. '-o bash' to display without format.",
    )
    @click.pass_context
    def env(ctx: click.Context, output: t.Literal["md", "bash"]) -> None:  # type: ignore (unused warning)
        if output not in ["md", "bash"]:
            raise CLIException(f"Unknown output format: {output}")

        is_windows = sys.platform == "win32"

        info_dict: dict[str, str | list[str]] = {
            "bentoml": importlib.metadata.version("bentoml"),
            "python": platform.python_version(),
            "platform": platform.platform(),
        }

        if is_windows:
            from ctypes import windll

            # https://stackoverflow.com/a/1026626
            is_admin: bool = windll.shell32.IsUserAnAdmin() != 0
            info_dict["is_window_admin"] = str(is_admin)
        else:
            info_dict["uid_gid"] = f"{os.geteuid()}:{os.getegid()}"

        if "CONDA_PREFIX" in os.environ:
            # conda packages
            conda_like = None
            for possible_exec in ["conda", "mamba", "micromamba"]:
                if shutil.which(possible_exec) is not None:
                    conda_like = possible_exec
                    break
            assert (
                conda_like is not None
            ), "couldn't find a conda-like executable, while CONDA_PREFIX is set."
            conda_packages = run_cmd([conda_like, "env", "export"])

            # user is currently in a conda environment,
            # doing this is faster than invoking `conda --version`
            try:
                conda_version = get_pkg_version("conda")
            except PackageNotFoundError:
                conda_version = run_cmd([conda_like, "--version"])[0].split(" ")[-1]

            info_dict[conda_like] = conda_version
            info_dict["in_conda_env"] = str(True)
            info_dict["conda_packages"] = conda_packages

        # process info from `pip freeze`
        pip_packages = run_cmd([sys.executable, "-m", "pip", "freeze"])
        info_dict["pip_packages"] = pip_packages
        click.echo(pretty_format(info_dict, output=output))
        ctx.exit(0)



from __future__ import annotations

import logging
import os
import subprocess
import typing as t
from abc import ABC
from abc import abstractmethod
from shutil import which
from tempfile import NamedTemporaryFile

import attr

from ...exceptions import BentoMLException
from ..bento.bento import Bento
from ..configuration import get_debug_mode

logger = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from fs.base import FS


@attr.define
class Environment(ABC):
    name: str
    env_fs: FS
    # path to bento's /env dir
    bento: Bento
    env_exe: str = attr.field(init=False)

    def __attrs_post_init__(self):
        self.env_exe = self.get_executable()
        if self.env_fs.exists(self.name):
            return self.env_fs.getsyspath(self.name)
        else:
            self.create()

    @abstractmethod
    def get_executable(self) -> str:
        """
        Returns executable path responsible for running this environment.

        Make sure that said executable is available in PATH.
        """
        ...

    @abstractmethod
    def create(self):
        """
        Create the environment with the files from bento.
        """
        ...

    @abstractmethod
    def run(self, commands: list[str]):
        """
        run the commands in an activated environment.
        """
        ...

    @staticmethod
    def run_script_subprocess(
        script_file_path: t.Union[str, os.PathLike[str]],
        capture_output: bool,
        debug_mode: bool,
    ):
        shell_path = which("bash")
        if shell_path is None:
            raise BentoMLException("Unable to locate a valid shell")

        safer_bash_args: list[str] = []

        # but only work in bash
        if debug_mode:
            safer_bash_args = ["-euxo", "pipefail"]
        result = subprocess.run(
            [shell_path, *safer_bash_args, script_file_path],
            capture_output=capture_output,
        )
        if result.returncode != 0:
            if result.stdout:
                logger.debug(result.stdout.decode())
            if result.stderr:
                logger.error(result.stderr.decode())
            raise BentoMLException(
                "Subprocess call returned non-zero value. Reffer logs for more details"
            )


class Conda(Environment):
    def get_executable(self) -> str:
        conda_exe = os.environ.get("CONDA_EXE")
        if conda_exe is None:
            raise BentoMLException(
                "Conda|Miniconda executable not found! Make sure any one is installed and environment is activated."
            )
        return conda_exe

    def create(self):
        # install conda deps
        from ..bento.build_config import CONDA_ENV_YAML_FILE_NAME

        # create a env under $BENTOML_HOME/env
        # setup conda with bento's environment.yml file and python/install.sh file
        conda_env_path = self.env_fs.getsyspath(self.name)
        python_version: str
        with open(self.bento.path_of("/env/python/version.txt"), "r") as pyver_file:
            py_version = pyver_file.read().split(".")[:2]
            python_version = ".".join(py_version)
        conda_environment_file = self.bento.path_of(
            f"/env/conda/{CONDA_ENV_YAML_FILE_NAME}"
        )
        create_script = f"""\
eval "$({self.get_executable()} shell.posix hook)"

conda create -p {conda_env_path} python={python_version} --yes

if [ -f {conda_environment_file} ]; then
    conda config --set pip_interop_enabled True
    conda env update -p {conda_env_path} --file {conda_environment_file}
fi

conda activate {conda_env_path}
bash -euxo pipefail {self.bento.path_of("/env/python/install.sh")}
"""
        with NamedTemporaryFile(mode="w", delete=False) as script_file:
            script_file.write(create_script)

        logger.info("Creating Conda env and installing dependencies...")
        self.run_script_subprocess(
            script_file.name,
            capture_output=not get_debug_mode(),
            debug_mode=get_debug_mode(),
        )

    def run(self, commands: list[str]):
        """
        Run commands in the activated environment.
        """
        conda_env_path = self.env_fs.getsyspath(self.name)
        create_script = f"""\
eval "$({self.env_exe} shell.posix hook)"
conda activate {conda_env_path}
{" ".join(commands)}
"""
        with NamedTemporaryFile(mode="w", delete=False) as script_file:
            script_file.write(create_script)
        self.run_script_subprocess(
            script_file.name, capture_output=False, debug_mode=get_debug_mode()
        )



from __future__ import annotations

import itertools
import logging
import shutil
import subprocess
import sys
import tempfile
import typing as t
from functools import partial

if t.TYPE_CHECKING:
    from click import Command
    from click import Context
    from click import Group
    from click import Parameter

    from bentoml._internal.container import DefaultBuilder

    from .utils import ClickParamType

    P = t.ParamSpec("P")
    F = t.Callable[P, t.Any]

logger = logging.getLogger("bentoml")


def compatible_option(*param_decls: str, **attrs: t.Any):
    """
    Make a new ``@click.option`` decorator that provides backwards compatibility.

    This decorator extends the default ``@click.option`` decorator by adding three
    additional parameters:

        * ``equivalent``: a tuple of (new_option, example_usage) that specifies
                          the new option and its example usage that replaces the given option.
        * ``append_msg``: a string to append a help message to the original help message.
    """
    import click

    from .utils import MEMO_KEY
    from .utils import normalize_none_type

    append_msg = attrs.pop("append_msg", "")
    equivalent: tuple[str, str] = attrs.pop("equivalent", ())
    factory = attrs.pop("factory", click)

    assert (
        equivalent is not None and len(equivalent) == 2
    ), "'equivalent' must be a tuple of (new_option, usage)"

    prepend_msg = "(Equivalent to ``--%s %s``):" % (*equivalent,)

    def obsolete_callback(ctx: Context, param: Parameter, value: ClickParamType):
        # NOTE: We want to transform memoized options to tuple[str] | bool | None

        assert param.name is not None, "argument name must be set."

        opt = param.opts[0]
        # ensure that our memoized are available under CLI context.
        if MEMO_KEY not in ctx.params:
            # By default, multiple options are stored as a tuple.
            # our memoized options are stored as a dict.
            ctx.params[MEMO_KEY] = {}

        # default can be a callable. We only care about the result.
        default_value = param.get_default(ctx)

        # We need to run transformation here such that we don't have to deal with
        # nested value.
        # NOTE: if users are using both old and new options, the new option will take
        # precedence, and the old option will be ignored.
        value = normalize_none_type(value)

        # if given param.name is not in the memoized options, we need to create them.
        if param.name not in ctx.params[MEMO_KEY]:
            # Initialize the memoized options with default value.
            ctx.params[MEMO_KEY][param.name] = () if param.multiple else default_value
            if value is not None:
                # Only warning if given value is different from the default.
                # Since we are going to transform default value from old options to
                # the kwargs map, there is no need to bloat logs.
                if value != default_value:
                    msg = "is now deprecated, use the equivalent"
                    if isinstance(value, bool):
                        logger.warning(
                            "'%s' %s '--%s %s' instead.",
                            opt,
                            msg,
                            equivalent[0],
                            opt.lstrip("--"),
                        )
                    elif isinstance(value, tuple):
                        obsolete_format = " ".join(
                            map(lambda s: "%s=%s" % (opt, s), value)
                        )
                        new_format = " ".join(
                            map(
                                lambda s: "--%s %s=%s"
                                % (equivalent[0], opt.lstrip("--"), s),
                                value,
                            )
                        )
                        logger.warning(
                            "'%s' %s '%s' instead.", obsolete_format, msg, new_format
                        )
                    else:
                        assert isinstance(value, str)
                        logger.warning(
                            "'%s=%s' %s '--%s %s=%s' instead.",
                            opt,
                            value,
                            msg,
                            equivalent[0],
                            opt.lstrip("--"),
                            value,
                        )
                if isinstance(value, (bool, str)):
                    ctx.params[MEMO_KEY][param.name] = value
                elif isinstance(value, tuple):
                    assert param.multiple
                    ctx.params[MEMO_KEY][param.name] += value
                else:
                    raise ValueError(f"Unexpected value type: {type(value)}")

    def decorator(f: F[t.Any]) -> t.Callable[[F[t.Any]], Command]:
        # We will set default value to None if 'default' is not set.
        msg = attrs.pop("help", None)
        assert msg is not None, "'help' is required."
        attrs.setdefault("help", " ".join([prepend_msg, msg, append_msg]))
        attrs.setdefault("expose_value", True)
        attrs.setdefault("callback", obsolete_callback)
        return factory.option(*param_decls, **attrs)(f)

    return decorator


def compatible_options_group(f: F[t.Any]):
    import click
    from click_option_group import optgroup

    optgroup_option = partial(compatible_option, factory=optgroup)

    compat = [
        optgroup_option(
            "--add-host",
            equivalent=("opt", "add-host=host:ip"),
            multiple=True,
            metavar="[HOST:IP,]",
            help="Add a custom host-to-IP mapping.",
        ),
        optgroup_option(
            "--build-arg",
            equivalent=("opt", "build-arg=FOO=bar"),
            multiple=True,
            metavar="[KEY=VALUE,]",
            help="Set build-time variables.",
        ),
        optgroup_option(
            "--cache-from",
            equivalent=("opt", "cache-from=user/app:cache"),
            multiple=True,
            default=None,
            metavar="[NAME|type=TYPE[,KEY=VALUE]]",
            help="External cache sources (e.g.: ``type=local,src=path/to/dir``).",
        ),
        optgroup_option(
            "--iidfile",
            type=click.STRING,
            equivalent=("opt", "iidfile=/path/to/file"),
            default=None,
            metavar="PATH",
            help="Write the image ID to the file.",
        ),
        optgroup_option(
            "--label",
            equivalent=("opt", "label=io.container.capabilities=baz"),
            multiple=True,
            metavar="[NAME|[KEY=VALUE],]",
            help="Set metadata for an image. This is equivalent to LABEL in given Dockerfile.",
        ),
        optgroup_option(
            "--network",
            type=click.STRING,
            equivalent=("opt", "network=default"),
            metavar="default|VALUE",
            default=None,
            help="Set the networking mode for the ``RUN`` instructions during build (default ``default``).",
        ),
        optgroup_option(
            "--no-cache",
            equivalent=("opt", "no-cache"),
            is_flag=True,
            default=False,
            help="Do not use cache when building the image.",
        ),
        optgroup_option(
            "--output",
            equivalent=("opt", "output=type=local,dest=/path/to/dir"),
            multiple=True,
            default=None,
            metavar="[PATH,-,type=TYPE[,KEY=VALUE]",
            help="Output destination.",
        ),
        optgroup_option(
            "--platform",
            equivalent=("opt", "platform=linux/amd64,linux/arm/v7"),
            default=None,
            metavar="VALUE[,VALUE]",
            multiple=True,
            type=click.STRING,
            help="Set target platform for build.",
        ),
        optgroup_option(
            "--progress",
            equivalent=("opt", "progress=auto"),
            default=None,
            type=click.Choice(["auto", "tty", "plain"]),
            metavar="VALUE",
            help="Set type of progress output. Use plain to show container output.",
        ),
        optgroup_option(
            "--pull",
            equivalent=("opt", "pull"),
            is_flag=True,
            default=False,
            help="Always attempt to pull all referenced images.",
        ),
        optgroup_option(
            "--secret",
            multiple=True,
            equivalent=("opt", "secret=id=aws,src=$HOME/.aws/crendentials"),
            metavar="[type=TYPE[,KEY=VALUE]",
            default=None,
            help="Secret to expose to the build.",
        ),
        optgroup_option(
            "--ssh",
            type=click.STRING,
            equivalent=("opt", "ssh=default=$SSH_AUTH_SOCK"),
            metavar="default|<id>[=<socket>|<key>[,<key>]]",
            default=None,
            help="SSH agent socket or keys to expose to the build. See https://docs.docker.com/engine/reference/commandline/buildx_build/#ssh",
        ),
        optgroup_option(
            "--target",
            equivalent=("opt", "target=release-stage"),
            metavar="VALUE",
            type=click.STRING,
            default=None,
            help="Set the target build stage to build.",
        ),
    ]
    for options in reversed(compat):
        f = options(f)
    return f


def buildx_options_group(f: F[t.Any]):
    import click
    from click_option_group import optgroup

    optgroup_option = partial(compatible_option, factory=optgroup)

    buildx = [
        optgroup_option(
            "--allow",
            equivalent=("opt", "allow=network.host"),
            multiple=True,
            metavar="ENTITLEMENT",
            help="Allow extra privileged entitlement (e.g., ``network.host``, ``security.insecure``).",
        ),
        optgroup_option(
            "--attest",
            equivalent=("opt", "attest=type=sbom,generator=image"),
            multiple=True,
            metavar="[NAME|type=TYPE[,KEY=VALUE]]",
            help="Attestation parameter (e.g., ``type=local,ref=path/to/dir``).",
        ),
        optgroup_option(
            "--build-context",
            equivalent=("opt", "build-context=project=path/to/project/source"),
            multiple=True,
            metavar="name=VALUE",
            help="Additional build contexts (e.g., ``name=path``).",
        ),
        optgroup_option(
            "--builder",
            equivalent=("opt", "builder=default"),
            type=click.STRING,
            help="Override the configured builder instance. Same as ``docker buildx --builder``",
        ),
        optgroup_option(
            "--cache-to",
            multiple=True,
            equivalent=("opt", "cache-to=type=registry,ref=user/app"),
            metavar="[NAME|type=TYPE[,KEY=VALUE]]",
            help="Cache export destinations (e.g., ``user/app:cache``, ``type=local,dest=path/to/dir``).",
        ),
        optgroup_option(
            "--load",
            is_flag=True,
            equivalent=("opt", "load"),
            default=False,
            help="Shorthand for ``--output=type=docker``. Note that ``--push`` and ``--load`` are mutually exclusive.",
        ),
        optgroup_option(
            "--push",
            is_flag=True,
            equivalent=("opt", "push"),
            default=False,
            help="Shorthand for ``--output=type=registry``. Note that ``--push`` and ``--load`` are mutually exclusive.",
        ),
        optgroup_option(
            "--provenance",
            equivalent=("opt", "provenance=generator=image"),
            type=click.STRING,
            help="Shorthand for ``--attest=type=provenance``.",
        ),
        optgroup_option(
            "--quiet",
            is_flag=True,
            equivalent=("opt", "quiet"),
            default=False,
            help="Suppress the build output and print image ID on success",
        ),
        optgroup_option(
            "--cgroup-parent",
            equivalent=("opt", "cgroup-parent=cgroupv2"),
            type=click.STRING,
            default=None,
            help="Optional parent cgroup for the container.",
        ),
        optgroup_option(
            "--ulimit",
            equivalent=("opt", "ulimit=nofile=1024:1024"),
            type=click.STRING,
            metavar="<type>=<soft limit>[:<hard limit>]",
            default=None,
            help="Ulimit options.",
        ),
        optgroup_option(
            "--no-cache-filter",
            equivalent=("opt", "no-cache-filter"),
            multiple=True,
            help="Do not cache specified stages.",
        ),
        optgroup_option(
            "--metadata-file",
            type=click.STRING,
            equivalent=("opt", "metadata-file=/path/to/file"),
            help="Write build result metadata to the file.",
        ),
        optgroup_option(
            "--sbom",
            equivalent=("opt", "sbom=generator=image"),
            type=click.STRING,
            help="Shorthand for ``--attest=type=sbom``.",
        ),
        optgroup_option(
            "--shm-size",
            equivalent=("opt", "shm-size=8192Mb"),
            help="Size of ``/dev/shm``.",
        ),
    ]
    for options in reversed(buildx):
        f = options(f)
    return f


def add_containerize_command(cli: Group) -> None:
    import click
    from click_option_group import optgroup

    from bentoml import container
    from bentoml._internal.configuration import get_debug_mode
    from bentoml._internal.container import FEATURES
    from bentoml._internal.container import REGISTERED_BACKENDS
    from bentoml._internal.container import determine_container_tag
    from bentoml._internal.container import enable_buildkit
    from bentoml.exceptions import BentoMLException
    from bentoml_cli.utils import kwargs_transformers
    from bentoml_cli.utils import normalize_none_type
    from bentoml_cli.utils import opt_callback
    from bentoml_cli.utils import validate_container_tag

    @cli.command()
    @click.argument("bento_tag", type=click.STRING, metavar="BENTO:TAG")
    @click.option(
        "-t",
        "--image-tag",
        "--docker-image-tag",
        metavar="NAME:TAG",
        help="Name and optionally a tag (format: ``name:tag``) for building container, defaults to the bento tag.",
        required=False,
        callback=validate_container_tag,
        multiple=True,
    )
    @click.option(
        "--backend",
        help="Define builder backend. Available: {}".format(
            ", ".join(map(lambda s: f"``{s}``", REGISTERED_BACKENDS))
        ),
        required=False,
        default="docker",
        envvar="BENTOML_CONTAINERIZE_BACKEND",
        type=click.STRING,
    )
    @click.option(
        "--enable-features",
        multiple=True,
        nargs=1,
        metavar="FEATURE[,FEATURE]",
        help="Enable additional BentoML features. Available features are: {}.".format(
            ", ".join(map(lambda s: f"``{s}``", FEATURES))
        ),
    )
    @click.option(
        "--opt",
        help="Define options for given backend. (format: ``--opt target=foo --opt build-arg=foo=BAR --opt iidfile:/path/to/file --opt no-cache``)",
        required=False,
        multiple=True,
        callback=opt_callback,
        metavar="ARG=VALUE[,ARG=VALUE]",
    )
    @click.option(
        "--run-as-root",
        help="Whether to run backend with as root.",
        is_flag=True,
        required=False,
        default=False,
        type=bool,
    )
    @optgroup.group(
        "Equivalent options", help="Equivalent option group for backward compatibility"
    )
    @compatible_options_group
    @optgroup.group("Buildx options", help="[Only available with '--backend=buildx']")
    @buildx_options_group
    @kwargs_transformers(transformer=normalize_none_type)
    def containerize(  # type: ignore
        bento_tag: str,
        image_tag: tuple[str] | None,
        backend: DefaultBuilder,
        enable_features: tuple[str] | None,
        run_as_root: bool,
        _memoized: dict[str, t.Any],
        **kwargs: t.Any,  # pylint: disable=unused-argument
    ) -> None:
        """Containerizes given Bento into an OCI-compliant container, with any given OCI builder.

        \b

        ``BENTO`` is the target BentoService to be containerized, referenced by its name
        and version in format of name:version. For example: ``iris_classifier:v1.2.0``

        \b
        ``bentoml containerize`` command also supports the use of the ``latest`` tag
        which will automatically use the last built version of your Bento.

        \b
        You can provide a tag for the image built by Bento using the
        ``--image-tag`` flag.

        \b
        You can also prefixing the tag with a hostname for the repository you wish to push to.

        .. code-block:: bash

           $ bentoml containerize iris_classifier:latest -t repo-address.com:username/iris

        This would build an image called ``username/iris:latest`` and push it to docker repository at ``repo-address.com``.

        \b
        By default, the ``containerize`` command will use the current credentials provided by each backend implementation.

        \b
        ``bentoml containerize`` also leverage BuildKit features such as multi-node
        builds for cross-platform images, more efficient caching, and more secure builds.
        BuildKit will be enabled by default by each backend implementation. To opt-out of BuildKit, set ``DOCKER_BUILDKIT=0``:

        .. code-block:: bash

            $ DOCKER_BUILDKIT=0 bentoml containerize iris_classifier:latest

        \b
        ``bentoml containerize`` is backend-agnostic, meaning that it can use any OCI-compliant builder to build the container image.
        By default, it will use the ``docker`` backend, which will use the local Docker daemon to build the image.
        You can use the ``--backend`` flag to specify a different backend. Note that backend flags will now be unified via ``--opt``.

        ``--opt`` accepts multiple flags, and it works with all backend options.

        \b
        To pass a boolean flag (e.g: ``--no-cache``), use ``--opt no-cache``.

        \b
        To pass a key-value pair (e.g: ``--build-arg FOO=BAR``), use ``--opt build-arg=FOO=BAR`` or ``--opt build-arg:FOO=BAR``. Make sure to quote the value if it contains spaces.

        \b
        To pass a argument with one value (path, integer, etc) (e.g: ``--cgroup-parent cgroupv2``), use ``--opt cgroup-parent=cgroupv2`` or ``--opt cgroup-parent:cgroupv2``.
        """

        # we can filter out all options that are None or False.
        _memoized = {k: v for k, v in _memoized.items() if v}
        logger.debug("Memoized: %s", _memoized)

        # Run healthcheck before containerizing
        # build will also run healthcheck, but we want to fail early.
        try:
            container.health(backend)
        except subprocess.CalledProcessError:
            raise BentoMLException(f"Backend {backend} is not healthy")

        # --progress is not available without BuildKit.
        if not enable_buildkit(backend=backend) and "progress" in _memoized:
            _memoized.pop("progress")
        elif get_debug_mode():
            _memoized["progress"] = "plain"
        _memoized["_run_as_root"] = run_as_root

        features: tuple[str] | None = None
        if enable_features:
            features = tuple(
                itertools.chain.from_iterable(
                    map(lambda s: s.split(","), enable_features)
                )
            )
        result = container.build(
            bento_tag,
            backend=backend,
            image_tag=image_tag,
            features=features,
            **_memoized,
        )

        if result is not None:
            tags = determine_container_tag(bento_tag, image_tag=image_tag)

            container_runtime = backend
            # This logic determines some of the output messages.
            if backend == "buildah":
                if shutil.which("podman") is None:
                    logger.warning(
                        "'buildah' are only used to build the image. To run the image, 'podman' is required to be installed (podman not found under PATH). See https://podman.io/getting-started/installation."
                    )
                    sys.exit(0)
                else:
                    container_runtime = "podman"
            elif backend == "buildx":
                # buildx is a syntatic sugar for docker buildx build
                container_runtime = "docker"
            elif backend == "buildctl":
                if shutil.which("docker") is not None:
                    container_runtime = "docker"
                elif shutil.which("podman") is not None:
                    container_runtime = "podman"
                elif shutil.which("nerdctl") is not None:
                    container_runtime = "nerdctl"
                else:
                    click.echo(
                        click.style(
                            "To load image built with 'buildctl' requires one of "
                            "docker, podman, nerdctl (None are found in PATH). "
                            "Make sure they are visible to PATH and try again.",
                            fg="yellow",
                        ),
                        color=True,
                    )
                    sys.exit(0)
                if "output" not in _memoized:
                    tmp_path = tempfile.gettempdir()
                    type_prefix = "type=oci,name="
                    if container_runtime == "docker":
                        type_prefix = "type=docker,name=docker.io/"
                    click.echo(
                        click.style(
                            "Autoconfig is now deprecated and will be removed "
                            "in the next future release. We recommend setting "
                            "output to a tarfile should you wish to load the "
                            "image locally. For example:",
                            fg="yellow",
                        ),
                        color=True,
                    )
                    click.echo(
                        click.style(
                            f"    bentoml containerize {bento_tag} --backend buildctl --opt output={type_prefix}{tags[0]},dest={tmp_path}/{tags[0].replace(':', '_')}.tar\n"
                            f"    {container_runtime} load -i {tmp_path}/{tags[0].replace(':', '_')}.tar",
                            fg="yellow",
                        ),
                        color=True,
                    )
                    o = subprocess.check_output(
                        [container_runtime, "load"], input=result
                    )
                    if get_debug_mode():
                        click.echo(o.decode("utf-8").strip())
                    sys.exit(0)
                return result

            click.echo(
                f'Successfully built Bento container for "{bento_tag}" with tag(s) "{",".join(tags)}"',
            )
            instructions = (
                "To run your newly built Bento container, run:\n"
                + f"    {container_runtime} run --rm -p 3000:3000 {tags[0]}\n"
            )

            if features is not None and any(
                has_grpc in features
                for has_grpc in ("all", "grpc", "grpc-reflection", "grpc-channelz")
            ):
                instructions += (
                    "To serve with gRPC instead, run:\n"
                    + f"    {container_runtime} run --rm -p 3000:3000 -p 3001:3001 {tags[0]} serve-grpc\n"
                )
            click.echo(instructions, nl=False)
            raise SystemExit(0)
        raise SystemExit(1)



from __future__ import annotations

import importlib.metadata
import logging
import os
import re
import typing as t
from functools import lru_cache

from ...exceptions import BentoMLConfigException
from ...exceptions import BentoMLException

# Find BentoML version managed by setuptools_scm
BENTOML_VERSION = importlib.metadata.version("bentoml")

if t.TYPE_CHECKING:
    from .containers import SerializationStrategy

# Note this file is loaded prior to logging being configured, thus logger is only
# used within functions in this file
logger = logging.getLogger(__name__)

DEBUG_ENV_VAR = "BENTOML_DEBUG"
QUIET_ENV_VAR = "BENTOML_QUIET"
CONFIG_ENV_VAR = "BENTOML_CONFIG"
CONFIG_OVERRIDE_ENV_VAR = "BENTOML_CONFIG_OPTIONS"
# https://github.com/grpc/grpc/blob/master/doc/environment_variables.md
GRPC_DEBUG_ENV_VAR = "GRPC_VERBOSITY"


def expand_env_var(env_var: str) -> str:
    """Expands potentially nested env var by repeatedly applying `expandvars` and
    `expanduser` until interpolation stops having any effect.
    """
    while True:
        interpolated = os.path.expanduser(os.path.expandvars(str(env_var)))
        if interpolated == env_var:
            return interpolated
        else:
            env_var = interpolated


def clean_bentoml_version(bentoml_version: str) -> str:
    post_version = bentoml_version.split("+")[0]
    match = re.match(r"^(\d+)\.(\d+)\.(\d+)(?:(a|rc)\d)*", post_version)
    if match is None:
        raise BentoMLException("Errors while parsing BentoML version.")
    return match.group()


@lru_cache(maxsize=1)
def is_pypi_installed_bentoml() -> bool:
    """Returns true if BentoML is installed via PyPI official release or installed from
     source with a release tag, which should come with pre-built docker base image on
     dockerhub.

    BentoML uses setuptools_scm to manage its versions, it looks at three things:

    * the latest tag (with a version number)
    * the distance to this tag (e.g. number of revisions since latest tag)
    * workdir state (e.g. uncommitted changes since latest tag)

    BentoML uses setuptools_scm with `version_scheme = "post-release"` option, which
    uses roughly the following logic to render the version:

    * no distance and clean: {tag}
    * distance and clean: {tag}.post{distance}+{scm letter}{revision hash}
    * no distance and not clean: {tag}+dYYYYMMDD
    * distance and not clean: {tag}.post{distance}+{scm letter}{revision hash}.dYYYYMMDD

    This function looks at the version str and decide if BentoML installation is
    base on a recent official release.
    """
    # In a git repo with no tag, setuptools_scm generated version starts with "0.1."
    try:
        from ..._version import __version_tuple__
    except ImportError:
        __version_tuple__ = (0, 0, 0, "dirty")

    is_tagged = not BENTOML_VERSION.startswith("0.1.")
    is_clean = not str(__version_tuple__[-1]).split(".")[-1].startswith("d")
    not_been_modified = BENTOML_VERSION == BENTOML_VERSION.split("+")[0]
    return is_tagged and is_clean and not_been_modified


def get_bentoml_config_file_from_env() -> str | None:
    if CONFIG_ENV_VAR in os.environ:
        # User local config file for customizing bentoml
        return expand_env_var(os.environ.get(CONFIG_ENV_VAR, ""))
    return None


def get_bentoml_override_config_from_env() -> str | None:
    if CONFIG_OVERRIDE_ENV_VAR in os.environ:
        # User local config options for customizing bentoml
        return os.environ.get(CONFIG_OVERRIDE_ENV_VAR, None)
    return None


def set_debug_mode(enabled: bool) -> None:
    os.environ[DEBUG_ENV_VAR] = str(enabled)
    os.environ[GRPC_DEBUG_ENV_VAR] = "DEBUG"

    logger.info(
        "%s debug mode for current BentoML session",
        "Enabling" if enabled else "Disabling",
    )


def get_debug_mode() -> bool:
    if DEBUG_ENV_VAR in os.environ:
        return os.environ[DEBUG_ENV_VAR].lower() == "true"
    return False


def set_quiet_mode(enabled: bool) -> None:
    # do not log setting quiet mode
    os.environ[QUIET_ENV_VAR] = str(enabled)
    os.environ[GRPC_DEBUG_ENV_VAR] = "NONE"


def get_quiet_mode() -> bool:
    if QUIET_ENV_VAR in os.environ:
        return os.environ[QUIET_ENV_VAR].lower() == "true"
    return False


def load_config(bentoml_config_file: str | None = None):
    """Load global configuration of BentoML"""

    from .containers import BentoMLConfiguration
    from .containers import BentoMLContainer

    if not bentoml_config_file:
        bentoml_config_file = get_bentoml_config_file_from_env()

    if bentoml_config_file:
        if not bentoml_config_file.endswith((".yml", ".yaml")):
            raise BentoMLConfigException(
                f"BentoML config file specified in ENV VAR does not end with either '.yaml' or '.yml': 'BENTOML_CONFIG={bentoml_config_file}'"
            ) from None
        if not os.path.isfile(bentoml_config_file):
            raise FileNotFoundError(
                f"BentoML config file specified in ENV VAR not found: 'BENTOML_CONFIG={bentoml_config_file}'"
            ) from None

    BentoMLContainer.config.set(
        BentoMLConfiguration(
            override_config_file=bentoml_config_file,
            override_config_values=get_bentoml_override_config_from_env(),
        ).to_dict()
    )


def save_config(config_file_handle: t.IO[t.Any]):
    import yaml

    from ..configuration.containers import BentoMLContainer

    content = yaml.safe_dump(BentoMLContainer.config)
    config_file_handle.write(content)


def set_serialization_strategy(serialization_strategy: SerializationStrategy):
    """
    Configure how bentoml.Service are serialized (pickled). Default via EXPORT_BENTO

    Args:
        serialization_strategy: One of ["EXPORT_BENTO", "LOCAL_BENTO", "REMOTE_BENTO"]
            EXPORT_BENTO: export and compress all Bento content as bytes
            LOCAL_BENTO: load Bento by tag in other processes, assuming available in local BentoStore
            REMOTE_BENTO: load Bento by tag in other processes, pull from Yatai/BentoCloud if not available locally
    """
    from ..configuration.containers import BentoMLContainer

    BentoMLContainer.serialization_strategy.set(serialization_strategy)



from ._internal.frameworks.easyocr import get
from ._internal.frameworks.easyocr import get_runnable
from ._internal.frameworks.easyocr import load_model
from ._internal.frameworks.easyocr import save_model

__all__ = ["load_model", "save_model", "get", "get_runnable"]



from __future__ import annotations

import io
import logging
from urllib.parse import urljoin

import requests

from ...exceptions import CloudRESTApiClientError
from ..configuration import BENTOML_VERSION
from .schemas import BentoRepositorySchema
from .schemas import BentoSchema
from .schemas import BentoWithRepositoryListSchema
from .schemas import ClusterFullSchema
from .schemas import ClusterListSchema
from .schemas import CompleteMultipartUploadSchema
from .schemas import CreateBentoRepositorySchema
from .schemas import CreateBentoSchema
from .schemas import CreateDeploymentSchema
from .schemas import CreateModelRepositorySchema
from .schemas import CreateModelSchema
from .schemas import DeploymentListSchema
from .schemas import DeploymentSchema
from .schemas import FinishUploadBentoSchema
from .schemas import FinishUploadModelSchema
from .schemas import ModelRepositorySchema
from .schemas import ModelSchema
from .schemas import ModelWithRepositoryListSchema
from .schemas import OrganizationSchema
from .schemas import PreSignMultipartUploadUrlSchema
from .schemas import UpdateBentoSchema
from .schemas import UpdateDeploymentSchema
from .schemas import UserSchema
from .schemas import schema_from_json
from .schemas import schema_from_object
from .schemas import schema_to_json

logger = logging.getLogger(__name__)


class RestApiClient:
    def __init__(self, endpoint: str, api_token: str) -> None:
        self.endpoint = endpoint
        self.session = requests.Session()
        self.session.headers.update(
            {
                "X-YATAI-API-TOKEN": api_token,
                "Content-Type": "application/json",
                "X-Bentoml-Version": BENTOML_VERSION,
            }
        )

    def _is_not_found(self, resp: requests.Response) -> bool:
        # We used to return 400 for record not found, handle both cases
        return (
            resp.status_code == 404
            or resp.status_code == 400
            and "record not found" in resp.text
        )

    def _check_resp(self, resp: requests.Response) -> None:
        if resp.status_code != 200:
            raise CloudRESTApiClientError(
                f"request failed with status code {resp.status_code}: {resp.text}"
            )

    def get_current_user(self) -> UserSchema | None:
        url = urljoin(self.endpoint, "/api/v1/auth/current")
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, UserSchema)

    def get_current_organization(self) -> OrganizationSchema | None:
        url = urljoin(self.endpoint, "/api/v1/current_org")
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, OrganizationSchema)

    def get_bento_repository(
        self, bento_repository_name: str
    ) -> BentoRepositorySchema | None:
        url = urljoin(
            self.endpoint, f"/api/v1/bento_repositories/{bento_repository_name}"
        )
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoRepositorySchema)

    def create_bento_repository(
        self, req: CreateBentoRepositorySchema
    ) -> BentoRepositorySchema:
        url = urljoin(self.endpoint, "/api/v1/bento_repositories")
        resp = self.session.post(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoRepositorySchema)

    def get_bento(self, bento_repository_name: str, version: str) -> BentoSchema | None:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}",
        )
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def create_bento(
        self, bento_repository_name: str, req: CreateBentoSchema
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint, f"/api/v1/bento_repositories/{bento_repository_name}/bentos"
        )
        resp = self.session.post(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def update_bento(
        self, bento_repository_name: str, version: str, req: UpdateBentoSchema
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}",
        )
        resp = self.session.patch(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def presign_bento_upload_url(
        self, bento_repository_name: str, version: str
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/presign_upload_url",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def presign_bento_download_url(
        self, bento_repository_name: str, version: str
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/presign_download_url",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def start_bento_multipart_upload(
        self, bento_repository_name: str, version: str
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/start_multipart_upload",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def presign_bento_multipart_upload_url(
        self,
        bento_repository_name: str,
        version: str,
        req: PreSignMultipartUploadUrlSchema,
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/presign_multipart_upload_url",
        )
        resp = self.session.patch(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def complete_bento_multipart_upload(
        self,
        bento_repository_name: str,
        version: str,
        req: CompleteMultipartUploadSchema,
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/complete_multipart_upload",
        )
        resp = self.session.patch(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def start_upload_bento(
        self, bento_repository_name: str, version: str
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/start_upload",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def finish_upload_bento(
        self, bento_repository_name: str, version: str, req: FinishUploadBentoSchema
    ) -> BentoSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/finish_upload",
        )
        resp = self.session.patch(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoSchema)

    def upload_bento(
        self, bento_repository_name: str, version: str, data: io.BytesIO
    ) -> None:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/upload",
        )
        resp = self.session.put(
            url,
            data=data,
            headers=dict(
                self.session.headers, **{"Content-Type": "application/octet-stream"}
            ),
        )
        self._check_resp(resp)
        return None

    def download_bento(
        self, bento_repository_name: str, version: str
    ) -> requests.Response:
        url = urljoin(
            self.endpoint,
            f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/download",
        )
        resp = self.session.get(url, stream=True)
        self._check_resp(resp)
        return resp

    def get_model_repository(
        self, model_repository_name: str
    ) -> ModelRepositorySchema | None:
        url = urljoin(
            self.endpoint, f"/api/v1/model_repositories/{model_repository_name}"
        )
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelRepositorySchema)

    def create_model_repository(
        self, req: CreateModelRepositorySchema
    ) -> ModelRepositorySchema:
        url = urljoin(self.endpoint, "/api/v1/model_repositories")
        resp = self.session.post(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelRepositorySchema)

    def get_model(self, model_repository_name: str, version: str) -> ModelSchema | None:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}",
        )
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def create_model(
        self, model_repository_name: str, req: CreateModelSchema
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint, f"/api/v1/model_repositories/{model_repository_name}/models"
        )
        resp = self.session.post(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def presign_model_upload_url(
        self, model_repository_name: str, version: str
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/presign_upload_url",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def presign_model_download_url(
        self, model_repository_name: str, version: str
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/presign_download_url",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def start_model_multipart_upload(
        self, model_repository_name: str, version: str
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/start_multipart_upload",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def presign_model_multipart_upload_url(
        self,
        model_repository_name: str,
        version: str,
        req: PreSignMultipartUploadUrlSchema,
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/presign_multipart_upload_url",
        )
        resp = self.session.patch(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def complete_model_multipart_upload(
        self,
        model_repository_name: str,
        version: str,
        req: CompleteMultipartUploadSchema,
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/complete_multipart_upload",
        )
        resp = self.session.patch(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def start_upload_model(
        self, model_repository_name: str, version: str
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/start_upload",
        )
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def finish_upload_model(
        self, model_repository_name: str, version: str, req: FinishUploadModelSchema
    ) -> ModelSchema:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/finish_upload",
        )
        resp = self.session.patch(url, data=schema_to_json(req))
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelSchema)

    def upload_model(
        self, model_repository_name: str, version: str, data: io.BytesIO
    ) -> None:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/upload",
        )
        resp = self.session.put(
            url,
            data=data,
            headers=dict(
                self.session.headers, **{"Content-Type": "application/octet-stream"}
            ),
        )
        self._check_resp(resp)
        return None

    def download_model(
        self, model_repository_name: str, version: str
    ) -> requests.Response:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models/{version}/download",
        )
        resp = self.session.get(url, stream=True)
        self._check_resp(resp)
        return resp

    def get_bento_repositories_list(
        self, bento_repository_name: str
    ) -> BentoWithRepositoryListSchema | None:
        url = urljoin(self.endpoint, "/api/v1/bento_repositories")
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoWithRepositoryListSchema)

    def get_bentos_list(self) -> BentoWithRepositoryListSchema | None:
        url = urljoin(self.endpoint, "/api/v1/bentos")
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, BentoWithRepositoryListSchema)

    def get_models_list(self) -> ModelWithRepositoryListSchema | None:
        url = urljoin(self.endpoint, "/api/v1/models")
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, ModelWithRepositoryListSchema)

    def get_deployment_list(
        self, cluster_name: str, **params: str | int | None
    ) -> DeploymentListSchema | None:
        url = urljoin(self.endpoint, f"/api/v1/clusters/{cluster_name}/deployments")
        resp = self.session.get(url, params=params)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, DeploymentListSchema)

    def create_deployment(
        self, cluster_name: str, create_schema: CreateDeploymentSchema
    ) -> DeploymentSchema | None:
        url = urljoin(self.endpoint, f"/api/v1/clusters/{cluster_name}/deployments")
        resp = self.session.post(url, data=schema_to_json(create_schema))
        self._check_resp(resp)
        return schema_from_json(resp.text, DeploymentSchema)

    def get_deployment(
        self, cluster_name: str, kube_namespace: str, deployment_name: str
    ) -> DeploymentSchema | None:
        url = urljoin(
            self.endpoint,
            f"/api/v1/clusters/{cluster_name}/namespaces/{kube_namespace}/deployments/{deployment_name}",
        )
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, DeploymentSchema)

    def update_deployment(
        self,
        cluster_name: str,
        kube_namespace: str,
        deployment_name: str,
        update_schema: UpdateDeploymentSchema,
    ) -> DeploymentSchema | None:
        url = urljoin(
            self.endpoint,
            f"/api/v1/clusters/{cluster_name}/namespaces/{kube_namespace}/deployments/{deployment_name}",
        )
        resp = self.session.patch(url, data=schema_to_json(update_schema))
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, DeploymentSchema)

    def terminate_deployment(
        self, cluster_name: str, kube_namespace: str, deployment_name: str
    ) -> DeploymentSchema | None:
        url = urljoin(
            self.endpoint,
            f"/api/v1/clusters/{cluster_name}/namespaces/{kube_namespace}/deployments/{deployment_name}/terminate",
        )
        resp = self.session.post(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, DeploymentSchema)

    def delete_deployment(
        self, cluster_name: str, kube_namespace: str, deployment_name: str
    ) -> DeploymentSchema | None:
        url = urljoin(
            self.endpoint,
            f"/api/v1/clusters/{cluster_name}/namespaces/{kube_namespace}/deployments/{deployment_name}",
        )
        resp = self.session.delete(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, DeploymentSchema)

    def get_cluster_list(
        self, params: dict[str, str | int] | None = None
    ) -> ClusterListSchema | None:
        url = urljoin(self.endpoint, "/api/v1/clusters")
        resp = self.session.get(url, params=params)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, ClusterListSchema)

    def get_cluster(self, cluster_name: str) -> ClusterFullSchema | None:
        url = urljoin(self.endpoint, f"/api/v1/clusters/{cluster_name}")
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_json(resp.text, ClusterFullSchema)

    def get_latest_model(
        self, model_repository_name: str, query: str | None = None
    ) -> ModelSchema | None:
        url = urljoin(
            self.endpoint,
            f"/api/v1/model_repositories/{model_repository_name}/models",
        )
        params = {"start": 0, "count": 10}
        if query:
            params["q"] = query
        resp = self.session.get(url, params=params)
        self._check_resp(resp)
        models = resp.json()["items"]
        return schema_from_object(models[0], ModelSchema) if models else None



"""
Bento Client.
=============

See https://docs.bentoml.com/en/latest/guides/client.html for more information.

.. code-block:: python

    import bentoml

    client = bentoml.client.Client.from_url("localhost:3000")

    client.predict(np.array([[5.9, 3, 5.1, 1.8]]))
"""
from __future__ import annotations

from ._internal.client import Client
from ._internal.client.grpc import GrpcClient
from ._internal.client.http import HTTPClient

__all__ = ["Client", "HTTPClient", "GrpcClient"]



from __future__ import annotations

import logging
import typing as t
from typing import TYPE_CHECKING

from simple_di import Provide
from simple_di import inject

from ._internal.configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    from ._internal.server.metrics.prometheus import PrometheusClient

logger = logging.getLogger(__name__)

# NOTE: We have to set our docstring here due to the fact that
# we are lazy loading the metrics. This means that the docstring
# won't be discovered until the metrics is initialized.
# this won't work with help() or doocstring on Sphinx.
# While this is less than optimal, we will do this since 'bentoml.metrics'
# is a public API.
_MAKE_WSGI_APP_DOCSTRING = """\
Create a WSGI app which serves the metrics from a registry.

Returns:
    WSGIApp: A WSGI app which serves the metrics from a registry.
"""
_GENERATE_LATEST_DOCSTRING = """\
Returns metrics from the registry in latest text format as a string.

This function ensures that multiprocess is setup correctly.

Returns:
    str: Metrics in latest text format. Refer to `Exposition format <https://prometheus.io/docs/instrumenting/exposition_formats/#exposition-formats>`_ for details.
"""
_TEXT_STRING_TO_METRIC_DOCSTRING = """
Parse Prometheus text format from a unicode string.

Returns:
    Metric: A generator that yields `Metric <https://prometheus.io/docs/concepts/metric_types/>`_ objects.
"""
_HISTOGRAM_DOCSTRING = """\
A Histogram tracks the size and number of events in a given bucket.

Histograms are often used to aggregatable calculation of quantiles.
Some notable examples include measuring response latency, request size.

A quick example of a Histogram:

.. code-block:: python

    from bentoml.metrics import Histogram

    h = Histogram('request_size_bytes', 'Request size (bytes)')

    @svc.api(input=JSON(), output=JSON())
    def predict(input_data: dict[str, str]):
        h.observe(512)  # Observe 512 (bytes)
        ...

``observe()`` will observe for given amount of time.
Usually, this value are positive or zero. Negative values are accepted but will
prevent current versions of Prometheus from properly detecting counter resets in the `sum of observations <https://prometheus.io/docs/practices/histograms/#count-and-sum-of-observations>`_.

Histograms also provide ``time()``, which times a block of code or function, and observe for a given duration amount.
This function can also be used as a context manager.

.. tab-set::

    .. tab-item:: Example

        .. code-block:: python

            from bentoml.metrics import Histogram

            REQUEST_TIME = Histogram('response_latency_seconds', 'Response latency (seconds)')

            @REQUEST_TIME.time()
            def create_response(request):
                body = await request.json()
                return Response(body)

    .. tab-item:: Context Manager

        .. code-block:: python

            from bentoml.metrics import Histogram

            REQUEST_TIME = Histogram('response_latency_seconds', 'Response latency (seconds)')

            def create_response(request):
                body = await request.json()
                with REQUEST_TIME.time():
                    ...

The default buckets are intended to cover a typical web/rpc request from milliseconds to seconds.
See :ref:`configuration guides <guides/configuration:Configuration>` to see how to customize the buckets.

Args:
    name (str): The name of the metric.
    documentation (str): A documentation string.
    labelnames (tuple[str]): A tuple of strings specifying the label names for the metric. Defaults to ``()``.
    namespace (str): The namespace of the metric. Defaults to an empty string.
    subsystem (str): The subsystem of the metric. Defaults to an empty string.
    unit (str): The unit of the metric. Defaults to an empty string.
    buckets (list[float]): A list of float representing a bucket. Defaults to ``(.005, .01, .025, .05, .075, .1, .25, .5, .75, 1.0, 2.5, 5.0, 7.5, 10.0, INF)``.
"""
_COUNTER_DOCSTRING = """
A Counter tracks counts of events or running totals.

.. epigraph::

    It is a cumulative metric that represents a single `monotonically increasing counter <https://prometheus.io/docs/concepts/metric_types/#counter>`_ whose value can only increase or be reset to zero on restart.

Some notable examples include counting the number of requests served, tasks completed, or errors.

If you need to go down, uses :func:`bentoml.metrics.Gauge` instead.

A quick example of a Counter:

.. code-block:: python

    from bentoml.metrics import Counter

    c = Counter('failures', 'Total number of failures requests')

    @svc.api(input=JSON(), output=JSON())
    def predict(input_data: dict[str, str]):
        if input_data['fail']:
            c.inc()  # increment by 1 by default

``inc()`` can optionally pass in a ``exemplar``, which is a dictionary of keys and values, defined :github:`here <OpenObservability/OpenMetrics/blob/main/specification/OpenMetrics.md#exemplars>`.

``inc()`` can also increment by any given amount:

.. code-block:: python

    c.inc(2.1)

``count_exceptions()`` can be used as both a decorator and context manager to count exceptions raised.

.. tab-set::

    .. tab-item:: Decorator

        .. code-block:: python

            from bentoml.metrics import Counter

            c = Counter('failures', 'Total number of failures requests')

            @c.count_exceptions()
            @svc.api(input=JSON(), output=JSON())
            def predict(input_data: dict[str, str]):
                if input_data['acc'] < 0.5:
                    raise ValueError("Given data is not accurate.")

    .. tab-item:: Context Manager

        .. code-block:: python

            from bentoml.metrics import Histogram

            c = Counter('failures', 'Total number of failures requests')

            @svc.api(input=JSON(), output=JSON())
            def predict(input_data: dict[str, str]):
                with c.count_exceptions():
                    if input_data['acc'] < 0.5:
                        raise ValueError("Given data is not accurate.")
                with c.count_exceptions(RuntimeError):
                    if input_data['output'] is None:
                        raise RuntimeError("Given pre-processing logic is invalid")

``count_exceptions()`` will optionally take in an exception to only track specific exceptions.

.. code-block:: python

    ...
    with c.count_exceptions(RuntimeError):
        if input_data['output'] is None:
            raise RuntimeError("Given pre-processing logic is invalid")

Args:
    name (str): The name of the metric.
    documentation (str): A documentation string.
    labelnames (tuple[str]): A tuple of strings specifying the label names for the metric. Defaults to ``()``.
    namespace (str): The namespace of the metric. Defaults to an empty string.
    subsystem (str): The subsystem of the metric. Defaults to an empty string.
    unit (str): The unit of the metric. Defaults to an empty string.
"""
_SUMMARY_DOCSTRING = """
A Summary tracks the size and `samples observations (usually things like request durations and response sizes).`.

While it also provides a total count of observations and a sum of all observed values,
it calculates configurable quantiles over a sliding time window.

Notable examples include request latency and response size.

A quick example of a Summary:

.. code-block:: python

    from bentoml.metrics import Summary

    s = Summary('request_size_bytes', 'Request size (bytes)')

    @svc.api(input=JSON(), output=JSON())
    def predict(input_data: dict[str, str]):
        s.observe(512)  # Observe 512 (bytes)
        ...

``observe()`` will observe for given amount of time.
Usually, this value are positive or zero. Negative values are accepted but will
prevent current versions of Prometheus from properly detecting counter resets in the `sum of observations <https://prometheus.io/docs/practices/histograms/#count-and-sum-of-observations>`_.

Similar to :meth:`bentoml.metrics.Histogram`, ``time()`` can also be used as a decorator or context manager.

.. tab-set::

    .. tab-item:: Example

        .. code-block:: python

            from bentoml.metrics import Histogram

            s = Summary('response_latency_seconds', 'Response latency (seconds)')

            @s.time()
            def create_response(request):
                body = await request.json()
                return Response(body)

    .. tab-item:: Context Manager

        .. code-block:: python

            from bentoml.metrics import Histogram

            s = Summary('response_latency_seconds', 'Response latency (seconds)')

            def create_response(request):
                body = await request.json()
                with s.time():
                    ...

Args:
    name (str): The name of the metric.
    documentation (str): A documentation string.
    labelnames (tuple[str]): A tuple of strings specifying the label names for the metric. Defaults to ``()``.
    namespace (str): The namespace of the metric. Defaults to an empty string.
    subsystem (str): The subsystem of the metric. Defaults to an empty string.
    unit (str): The unit of the metric. Defaults to an empty string.
"""
_GAUGE_DOCSTRING = """
A Gauge represents a single numerical value that can arbitrarily go up and down.

Gauges are typically used to for report instantaneous values like temperatures or current memory usage.
One can think of Gauge as a :meth:`bentoml.metrics.Counter` that can go up and down.

Notable examples include in-progress requests, number of item in a queue, and free memory.

A quick example of a Gauge:

.. code-block:: python

    from bentoml.metrics import Gauge

    g = Gauge('inprogress_request', 'Request inprogress')

    @svc.api(input=JSON(), output=JSON())
    def predict(input_data: dict[str, str]):
        g.inc()  # increment by 1 by default
        g.dec(10) # decrement by any given value
        g.set(0)  # set to a given value
        ...

.. note::

    By default, ``inc()`` and ``dec()`` will increment and decrement by 1 respectively.

Gauge also provide ``track_inprogress()``, to track inprogress object.
This function can also be used as either a context manager or a decorator.

.. tab-set::

    .. tab-item:: Example

        .. code-block:: python

            from bentoml.metrics import Gauge

            g = Gauge('inprogress_request', 'Request inprogress')

            @svc.api(input=JSON(), output=JSON())
            @g.track_inprogress()
            def predict(input_data: dict[str, str]):
                ...

    .. tab-item:: Context Manager

        .. code-block:: python

            from bentoml.metrics import Gauge

            g = Gauge('inprogress_request', 'Request inprogress')

            @svc.api(input=JSON(), output=JSON())
            def predict(input_data: dict[str, str]):
                with g.track_inprogress():
                    ...

        The gauge will increment when the context is entered and decrement when the context is exited.

Args:
    name (str): The name of the metric.
    documentation (str): A documentation string.
    labelnames (tuple[str]): A tuple of strings specifying the label names for the metric. Defaults to ``()``.
    namespace (str): The namespace of the metric. Defaults to an empty string.
    subsystem (str): The subsystem of the metric. Defaults to an empty string.
    unit (str): The unit of the metric. Defaults to an empty string.
    multiprocess_mode (str): The multiprocess mode of the metric. Defaults to ``all``. Available options
                             are (``all``, ``min``, ``max``, ``livesum``, ``liveall``)
"""

# This sets of functions are implemented in the PrometheusClient class
_INTERNAL_FN_IMPL = {
    "make_wsgi_app": _MAKE_WSGI_APP_DOCSTRING,
    "generate_latest": _GENERATE_LATEST_DOCSTRING,
    "text_string_to_metric_families": _TEXT_STRING_TO_METRIC_DOCSTRING,
}
_NOT_IMPLEMENTED = [
    "delete_from_gateway",
    "instance_ip_grouping_key",
    "push_to_gateway",
    "pushadd_to_gateway",
]
_NOT_SUPPORTED = [
    "GC_COLLECTOR",
    "GCCollector",
    "PLATFORM_COLLECTOR",
    "PlatformCollector",
    "PROCESS_COLLECTOR",
    "ProcessCollector",
    "REGISTRY",
    "CONTENT_TYPE_LATEST",
    "start_http_server",
    "start_wsgi_server",
    "make_asgi_app",
    "write_to_textfile",
] + _NOT_IMPLEMENTED
_docstring = {
    "Counter": _COUNTER_DOCSTRING,
    "Histogram": _HISTOGRAM_DOCSTRING,
    "Summary": _SUMMARY_DOCSTRING,
    "Gauge": _GAUGE_DOCSTRING,
}
_docstring.update(_INTERNAL_FN_IMPL)


def __dir__() -> list[str]:
    # This is for IPython and IDE autocompletion.
    metrics_client = BentoMLContainer.metrics_client.get()
    return list(set(dir(metrics_client.prometheus_client)) - set(_NOT_SUPPORTED))


def __getattr__(item: t.Any):
    if item in _NOT_SUPPORTED:
        raise NotImplementedError(
            f"{item} is not supported when using '{__name__}'. See https://docs.bentoml.com/en/latest/reference/metrics.html."
        )
    # This is the entrypoint for all bentoml.metrics.*
    return _LazyMetric(item, docstring=_docstring.get(item))


class _LazyMetric:
    __slots__ = ("_attr", "_proxy", "_initialized", "_args", "_kwargs", "__doc__")

    def __init__(self, attr: str, docstring: str | None = None):
        self._attr = attr
        self.__doc__ = docstring
        self._proxy = None
        self._initialized = False
        self._args: tuple[t.Any, ...] = ()
        self._kwargs: dict[str, t.Any] = {}

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        """
        Lazily initialize the metrics object.

        Args:
            *args: Arguments to pass to the metrics object.
            **kwargs: Keyword arguments to pass to the metrics object.
        """
        if "registry" in kwargs:
            raise ValueError(
                f"'registry' should not be passed when using '{__name__}.{self._attr}'. See https://docs.bentoml.com/en/latest/reference/metrics.html."
            )
        self._args = args
        self._kwargs = kwargs
        if self._attr in _INTERNAL_FN_IMPL:
            # first-class function implementation from BentoML Prometheus client.
            # In this case, the function will be called directly.
            return self._load_proxy()
        return self

    def __getattr__(self, item: t.Any) -> t.Any:
        if item in self.__slots__:
            raise AttributeError(f"Attribute {item} is private to {self}.")
        if self._proxy is None:
            self._proxy = self._load_proxy()
        assert self._initialized and self._proxy is not None
        return getattr(self._proxy, item)

    def __dir__(self) -> list[str]:
        if self._proxy is None:
            self._proxy = self._load_proxy()
        assert self._initialized and self._proxy is not None
        return dir(self._proxy)

    @inject
    def _load_proxy(
        self,
        metrics_client: PrometheusClient = Provide[BentoMLContainer.metrics_client],
    ) -> None:
        client_impl = (
            metrics_client
            if self._attr in dir(metrics_client)
            else metrics_client.prometheus_client
        )
        proxy = getattr(client_impl, self._attr)(*self._args, **self._kwargs)
        self._initialized = True
        return proxy
